import json
import os
import unittest

import mock
from pyramid import testing

from kinto.core.cache import postgresql as postgresql_cache
from kinto.core.permission import postgresql as postgresql_permission
from kinto.core.storage import postgresql as postgresql_storage
from kinto.core.storage.postgresql.migrator import MigratorMixin
from kinto.core.testing import skip_if_no_postgresql


@mock.patch.object(MigratorMixin, '_execute_sql_file')
class MigratorTest(unittest.TestCase):
    def setUp(self):
        self.migrator = MigratorMixin()
        here = os.path.dirname(__file__)
        migrations_directory = os.path.join(here, 'migrations')
        self.migrator.migrations_directory = migrations_directory
        self.migrator.schema_version = 6

    def test_schema_is_created_if_no_version(self, execute_sql):
        with mock.patch.object(self.migrator, 'create_schema') as create_schema:
            with mock.patch.object(self.migrator, 'get_installed_version') as installed_version:
                installed_version.return_value = None
                self.migrator.create_or_migrate_schema()
        self.assertTrue(create_schema.called)

    def test_schema_is_not_touched_if_already_current(self, execute_sql):
        # Patch to keep track of SQL files executed.
        with mock.patch.object(self.migrator, 'get_installed_version') as installed_version:
            installed_version.return_value = 6
            self.migrator.create_or_migrate_schema()
        self.assertFalse(execute_sql.called)

    def _walk_from_3_to_6(self):
        """Helper method for simulating a migration from version 3 to version 6."""
        versions = [6, 5, 4, 3, 3]
        self.migrator.get_installed_version = lambda: versions.pop()

    def test_migration_file_is_executed_for_every_intermediary_version(self, execute_sql):
        self._walk_from_3_to_6()

        self.migrator.create_or_migrate_schema()
        sql_called = execute_sql.call_args_list[-3][0][0]
        self.assertIn('migrations/migration_003_004.sql', sql_called)
        sql_called = execute_sql.call_args_list[-2][0][0]
        self.assertIn('migrations/migration_004_005.sql', sql_called)
        sql_called = execute_sql.call_args_list[-1][0][0]
        self.assertIn('migrations/migration_005_006.sql', sql_called)

    def test_migration_files_are_listed_if_ran_with_dry_run(self, execute_sql):
        self._walk_from_3_to_6()

        with mock.patch('kinto.core.storage.postgresql.migrator.logger') as mocked:
            self.migrator.create_or_migrate_schema(dry_run=True)

        output = ''.join([repr(call) for call in mocked.info.call_args_list])
        self.assertIn('migrations/migration_003_004.sql', output)
        self.assertIn('migrations/migration_004_005.sql', output)
        self.assertIn('migrations/migration_005_006.sql', output)

    def test_migration_fails_if_intermediary_version_is_missing(self, execute_sql):
        with mock.patch.object(self.migrator,
                               'get_installed_version') as current:
            current.return_value = -1
            self.assertRaises(AssertionError, self.migrator.create_or_migrate_schema)


@skip_if_no_postgresql
class PostgresqlStorageMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_storage import PostgreSQLStorageTest
        self.settings = {**PostgreSQLStorageTest.settings}
        self.config = testing.setUp()
        self.config.add_settings(self.settings)
        self.version = postgresql_storage.Storage.schema_version
        # Usual storage object to manipulate the storage.
        self.storage = postgresql_storage.load_from_config(self.config)

    def setUp(self):
        # Start empty.
        self._delete_everything()
        # Create schema in its last version
        self.storage.initialize_schema()

    def tearDown(self):
        postgresql_storage.Storage.schema_version = self.version

    def _delete_everything(self):
        q = """
        DROP TABLE IF EXISTS records CASCADE;
        DROP TABLE IF EXISTS deleted CASCADE;
        DROP TABLE IF EXISTS metadata CASCADE;
        DROP FUNCTION IF EXISTS resource_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS collection_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS bump_timestamp();
        """
        with self.storage.client.connect() as conn:
            conn.execute(q)

    def _load_schema(self, filepath):
        with self.storage.client.connect() as conn:
            here = os.path.dirname(__file__)
            with open(os.path.join(here, filepath)) as f:
                old_schema = f.read()
            conn.execute(old_schema)

    def test_does_not_execute_if_ran_with_dry(self):
        self._delete_everything()
        self.storage.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'records';"""
        with self.storage.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)

    def test_schema_sets_the_current_version(self):
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

    def test_schema_is_considered_first_version_if_no_version_detected(self):
        with self.storage.client.connect() as conn:
            q = "DELETE FROM metadata WHERE name = 'storage_schema_version';"
            conn.execute(q)

        self.assertEqual(self.storage.get_installed_version(), 1)

    def test_schema_is_considered_20_if_server_is_wiped(self):
        with self.storage.client.connect() as conn:
            q = "DELETE FROM metadata;"
            conn.execute(q)

        self.assertEqual(self.storage.get_installed_version(), 20)

    def test_every_available_migration(self):
        """Test every migration available in kinto.core code base since
        version 1.6.

        Records migration test is currently very naive, and should be
        elaborated along future migrations.
        """
        self._delete_everything()

        # Install old schema
        self._load_schema('schema/postgresql-storage-1.6.sql')

        # Create a sample record using some code that is compatible with the
        # schema in place in cliquet 1.6.
        with self.storage.client.connect() as conn:
            before = {'drink': 'cacao'}
            query = """
            INSERT INTO records (user_id, resource_name, data)
            VALUES (:user_id, :resource_name, (:data)::JSON)
            RETURNING id, as_epoch(last_modified) AS last_modified;
            """
            placeholders = dict(user_id='jean-louis',
                                resource_name='test',
                                data=json.dumps(before))
            result = conn.execute(query, placeholders)
            inserted = result.fetchone()
            before['id'] = str(inserted['id'])
            before['last_modified'] = inserted['last_modified']

        # In cliquet 1.6, version = 1.
        version = self.storage.get_installed_version()
        self.assertEqual(version, 1)

        # Run every migrations available.
        self.storage.initialize_schema()

        # Version matches current one.
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

        # Check that previously created record is still here
        migrated, count = self.storage.get_all('test', 'jean-louis')
        self.assertEqual(migrated[0], before)

        # Check that new records can be created
        r = self.storage.create('test', ',jean-louis', {'drink': 'mate'})

        # And deleted
        self.storage.delete('test', ',jean-louis', r['id'])

    def test_every_available_migration_succeeds_if_tables_were_flushed(self):
        # During tests, tables can be flushed.
        self.storage.flush()
        self.storage.initialize_schema()
        # Version matches current one.
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

    def test_migration_12_clean_tombstones(self):
        self._delete_everything()
        last_version = postgresql_storage.Storage.schema_version
        postgresql_storage.Storage.schema_version = 11

        self._load_schema('schema/postgresql-storage-11.sql')

        insert_query = """
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:id, :parent_id, :collection_id, (:data)::JSONB, from_epoch(:last_modified))
        """
        placeholders = dict(id='rid',
                            parent_id='jean-louis',
                            collection_id='test',
                            data=json.dumps({'drink': 'mate'}),
                            last_modified=123456)
        with self.storage.client.connect() as conn:
            conn.execute(insert_query, placeholders)

        create_tombstone = """
        INSERT INTO deleted (id, parent_id, collection_id, last_modified)
        VALUES (:id, :parent_id, :collection_id, from_epoch(:last_modified))
        """
        with self.storage.client.connect() as conn:
            conn.execute(create_tombstone, placeholders)

        # Execute the 011 to 012 migration (and others)
        postgresql_storage.Storage.schema_version = last_version
        self.storage.initialize_schema()

        # Check that the rotted tombstone has been removed, but the
        # original record remains.
        records, count = self.storage.get_all('test', 'jean-louis')
        # Only the record remains.
        assert len(records) == 1
        assert count == 1

    def test_migration_18_merges_tombstones(self):
        self._delete_everything()
        last_version = postgresql_storage.Storage.schema_version

        self._load_schema('schema/postgresql-storage-11.sql')
        # Schema 11 is essentially the same as schema 17
        postgresql_storage.Storage.schema_version = 17
        with self.storage.client.connect() as conn:
            conn.execute("""
            UPDATE metadata SET value = '17'
            WHERE name = 'storage_schema_version';
            """)

        insert_query = """
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:id, :parent_id, :collection_id, (:data)::JSONB, from_epoch(:last_modified))
        """
        placeholders = dict(id='rid',
                            parent_id='jean-louis',
                            collection_id='test',
                            data=json.dumps({'drink': 'mate'}),
                            last_modified=123456)
        with self.storage.client.connect() as conn:
            conn.execute(insert_query, placeholders)

        create_tombstone = """
        INSERT INTO deleted (id, parent_id, collection_id, last_modified)
        VALUES (:id, :parent_id, :collection_id, from_epoch(:last_modified))
        """
        with self.storage.client.connect() as conn:
            conn.execute(create_tombstone, placeholders)

        # Execute the 017 to 018 migration (and others)
        postgresql_storage.Storage.schema_version = last_version
        self.storage.initialize_schema()

        # Check that the record took precedence of over the tombstone.
        records, count = self.storage.get_all('test', 'jean-louis',
                                              include_deleted=True)
        assert len(records) == 1
        assert count == 1
        assert records[0]['drink'] == 'mate'


@skip_if_no_postgresql
class PostgresqlPermissionMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_permission import PostgreSQLPermissionTest
        settings = {**PostgreSQLPermissionTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.permission = postgresql_permission.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS access_control_entries CASCADE;
        DROP TABLE IF EXISTS user_principals CASCADE;
        DROP TABLE IF EXISTS metadata CASCADE;
        """
        with self.permission.client.connect() as conn:
            conn.execute(q)

    def _load_schema(self, filepath):
        with self.permission.client.connect() as conn:
            here = os.path.dirname(__file__)
            with open(os.path.join(here, filepath)) as f:
                old_schema = f.read()
            conn.execute(old_schema)

    def test_assumes_schema_1_if_no_permission_schema_version(self):
        self.permission.initialize_schema()
        q = """
        DELETE FROM metadata WHERE name = 'permission_schema_version';
        """
        with self.permission.client.connect() as conn:
            conn.execute(q)

        self.assertEqual(self.permission.get_installed_version(), 1)

    def test_assumes_schema_1_if_no_metadata_table(self):
        self.permission.initialize_schema()
        q = """
        DROP TABLE metadata;
        """
        with self.permission.client.connect() as conn:
            conn.execute(q)

        self.assertEqual(self.permission.get_installed_version(), 1)

    def test_assumes_no_schema_if_no_user_principals_table(self):
        # Verify that the permission migrations assume no schema if
        # the user_principals table is missing, even if the metadata
        # table is present from the storage migrations.
        self.permission.initialize_schema()
        q = """
        DELETE FROM metadata WHERE name = 'permission_schema_version';
        DROP TABLE user_principals;
        """
        with self.permission.client.connect() as conn:
            conn.execute(q)

        self.assertEqual(self.permission.get_installed_version(), None)

    def test_schema_creation_leaves_at_max_version(self):
        self.permission.initialize_schema()
        self.assertEqual(self.permission.get_installed_version(), self.permission.schema_version)

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.permission.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(query)
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.permission.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)

    def test_every_available_migration(self):
        """Test every permission migration available in code base.

        Records migration test is currently very naive, and should be
        elaborated along future migrations.
        """
        # Install old schema
        self._load_schema('schema/postgresql-permission-1.sql')

        # Create some permissions data using some code that is
        # compatible with the schema in place.
        with self.permission.client.connect() as conn:
            query = """
            INSERT INTO user_principals
            VALUES ('remy', 'admin');
            INSERT INTO access_control_entries VALUES
              ('sailboat', 'write', 'remy'),
              ('sailboat', 'read', 'ethan'),
              ('sailboat/log', 'read', 'system.Authenticated'),
              ('sailboat/log', 'write', 'admin');
            """
            conn.execute(query)

        version = self.permission.get_installed_version()
        self.assertEqual(version, 1)

        # Run every migrations available.
        self.permission.initialize_schema()

        # Version matches current one.
        version = self.permission.get_installed_version()
        self.assertEqual(version, self.permission.schema_version)

        # Check that previously created data is still available
        remy_principals = self.permission.get_user_principals('remy')
        self.assertEqual(remy_principals, set(['admin']))

        remy_objects = self.permission.get_accessible_objects(
            ['remy', 'admin', 'system.Authenticated'])
        self.assertEqual(remy_objects, {
            'sailboat': set(['write']),
            'sailboat/log': set(['read', 'write']),
        })

        # Check that new records can be created
        self.permission.add_user_principal('ethan', 'crew')

        # And deleted
        self.permission.remove_principal_from_ace('sailboat/log', 'read', 'system.Authenticated')


@skip_if_no_postgresql
class PostgresqlCacheMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_cache import PostgreSQLCacheTest
        settings = {**PostgreSQLCacheTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.cache = postgresql_cache.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS cache CASCADE;
        """
        with self.cache.client.connect() as conn:
            conn.execute(q)

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.cache.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(query)
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.cache.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)


class PostgresqlExceptionRaisedTest(unittest.TestCase):
    def setUp(self):
        self.sqlalchemy = postgresql_storage.client.sqlalchemy

    def tearDown(self):
        postgresql_storage.client.sqlalchemy = self.sqlalchemy

    def test_postgresql_usage_raise_an_error_if_postgresql_not_installed(self):
        postgresql_storage.client.sqlalchemy = None
        with self.assertRaises(ImportWarning):
            postgresql_storage.client.create_from_config(testing.setUp())



import mock
import pytest

from kinto.core.utils import COMPARISON, sqlalchemy
from kinto.core.storage import generators, memory, postgresql, exceptions, StorageBase
from kinto.core.storage import Filter, Sort, MISSING
from kinto.core.storage.testing import StorageTest
from kinto.core.storage.utils import paginated
from kinto.core.testing import (unittest, skip_if_no_postgresql)


class GeneratorTest(unittest.TestCase):
    def test_generic_has_mandatory_override(self):
        self.assertRaises(NotImplementedError, generators.Generator)

    def test_id_generator_must_respect_storage_backends(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return '*' * 80

        self.assertRaises(ValueError, Dumb)

    def test_default_generator_allow_underscores_dash_alphabet(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return '1234'

        generator = Dumb()
        self.assertTrue(generator.match('1_2_3-abc'))
        self.assertTrue(generator.match('abc_123'))
        self.assertFalse(generator.match('-1_2_3-abc'))
        self.assertFalse(generator.match('_1_2_3-abc'))

    def test_uuid_generator_pattern_allows_uuid_only(self):
        invalid_uuid = 'XXX-00000000-0000-5000-a000-000000000000'
        generator = generators.UUID4()
        self.assertFalse(generator.match(invalid_uuid))

    def test_uuid_generator_pattern_is_not_restricted_to_uuid4(self):
        generator = generators.UUID4()
        valid_uuid = 'fd800e8d-e8e9-3cac-f502-816cbed9bb6c'
        self.assertTrue(generator.match(valid_uuid))
        invalid_uuid4 = '00000000-0000-5000-a000-000000000000'
        self.assertTrue(generator.match(invalid_uuid4))
        invalid_uuid4 = '00000000-0000-4000-e000-000000000000'
        self.assertTrue(generator.match(invalid_uuid4))


class StorageBaseTest(unittest.TestCase):
    def setUp(self):
        self.storage = StorageBase()

    def test_mandatory_overrides(self):
        calls = [
            (self.storage.initialize_schema,),
            (self.storage.flush,),
            (self.storage.collection_timestamp, '', ''),
            (self.storage.create, '', '', {}),
            (self.storage.get, '', '', ''),
            (self.storage.update, '', '', '', {}),
            (self.storage.delete, '', '', ''),
            (self.storage.delete_all, '', ''),
            (self.storage.purge_deleted, '', ''),
            (self.storage.get_all, '', ''),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)

    def test_backend_error_message_provides_given_message_if_defined(self):
        error = exceptions.BackendError(message="Connection Error")
        self.assertEqual(str(error), "Connection Error")

    def test_backenderror_message_default_to_original_exception_message(self):
        error = exceptions.BackendError(ValueError("Pool Error"))
        self.assertEqual(str(error), "ValueError: Pool Error")


class MemoryBasedStorageTest(unittest.TestCase):
    def test_backend_raise_not_implemented_error(self):
        storage = memory.MemoryBasedStorage()
        with pytest.raises(NotImplementedError):
            storage.bump_and_store_timestamp("record", "/buckets/foo/collections/bar")


class MemoryStorageTest(StorageTest, unittest.TestCase):
    backend = memory
    settings = {
        'storage_strict_json': True
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage,
            'bump_and_store_timestamp',
            side_effect=exceptions.BackendError("Segmentation fault."))

    def test_backend_error_provides_original_exception(self):
        pass

    def test_raises_backend_error_if_error_occurs_on_client(self):
        pass

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_backenderror_message_default_to_original_exception_message(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass


class LenientMemoryStorageTest(MemoryStorageTest):
    settings = {
        'storage_strict_json': False
    }

    def test_create_bytes_raises(self):
        data = {'steak': 'haché'.encode(encoding='utf-8')}
        self.assertIsInstance(data['steak'], bytes)
        self.assertIsNotNone(self.create_record(data))

    def test_update_bytes_raises(self):
        record = self.create_record()

        new_record = {'steak': 'haché'.encode(encoding='utf-8')}
        self.assertIsInstance(new_record['steak'], bytes)

        self.assertIsNotNone(self.storage.update(
                                object_id=record['id'],
                                record=new_record,
                                **self.storage_kw))


@skip_if_no_postgresql
class PostgreSQLStorageTest(StorageTest, unittest.TestCase):
    backend = postgresql
    settings = {
        'storage_max_fetch_size': 10000,
        'storage_backend': 'kinto.core.storage.postgresql',
        'storage_poolclass': 'sqlalchemy.pool.StaticPool',
        'storage_url': 'postgres://postgres:postgres@localhost:5432/testdb',
        'storage_strict_json': True
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage.client,
            'session_factory',
            side_effect=sqlalchemy.exc.SQLAlchemyError)

    def test_number_of_fetched_records_can_be_limited_in_settings(self):
        for i in range(4):
            self.create_record({'phone': 'tel-{}'.format(i)})

        results, count = self.storage.get_all(**self.storage_kw)
        self.assertEqual(len(results), 4)

        settings = {**self.settings, 'storage_max_fetch_size': 2}
        config = self._get_config(settings=settings)
        limited = self.backend.load_from_config(config)

        results, count = limited.get_all(**self.storage_kw)
        self.assertEqual(count, 4)
        self.assertEqual(len(results), 2)

    def test_number_of_fetched_records_is_per_page(self):
        for i in range(10):
            self.create_record({'number': i})

        settings = {**self.settings, 'storage_max_fetch_size': 2}
        config = self._get_config(settings=settings)
        backend = self.backend.load_from_config(config)

        results, count = backend.get_all(pagination_rules=[
                                             [Filter('number', 1, COMPARISON.GT)]
                                         ], **self.storage_kw)
        self.assertEqual(count, 10)
        self.assertEqual(len(results), 2)

    def test_connection_is_rolledback_if_error_occurs(self):
        with self.storage.client.connect() as conn:
            query = "DELETE FROM records WHERE collection_id = 'genre';"
            conn.execute(query)

        try:
            with self.storage.client.connect() as conn:
                query = """
                INSERT INTO records VALUES ('rock-and-roll', 'music', 'genre', NOW(), '{}', FALSE);
                """
                conn.execute(query)
                conn.commit()

                query = """
                INSERT INTO records VALUES ('jazz', 'music', 'genre', NOW(), '{}', FALSE);
                """
                conn.execute(query)

                raise sqlalchemy.exc.TimeoutError()
        except exceptions.BackendError:
            pass

        with self.storage.client.connect() as conn:
            query = "SELECT COUNT(*) FROM records WHERE collection_id = 'genre';"
            result = conn.execute(query)
            self.assertEqual(result.fetchone()[0], 1)

    def test_pool_object_is_shared_among_backend_instances(self):
        config = self._get_config()
        storage1 = self.backend.load_from_config(config)
        storage2 = self.backend.load_from_config(config)
        self.assertEqual(id(storage1.client),
                         id(storage2.client))

    def test_warns_if_configured_pool_size_differs_for_same_backend_type(self):
        self.backend.load_from_config(self._get_config())
        settings = {**self.settings, 'storage_pool_size': 1}
        msg = ('Reuse existing PostgreSQL connection. Parameters storage_* '
               'will be ignored.')
        with mock.patch('kinto.core.storage.postgresql.client.'
                        'warnings.warn') as mocked:
            self.backend.load_from_config(self._get_config(settings=settings))
            mocked.assert_any_call(msg)

    def test_get_all_raises_if_missing_on_strange_query(self):
        with self.assertRaises(ValueError):
            self.storage.get_all('some-collection', 'some-parent',
                                 filters=[Filter("author", MISSING, COMPARISON.HAS)])

    def test_integrity_error_rollsback_transaction(self):
        client = postgresql.create_from_config(self._get_config(),
                                               prefix='storage_',
                                               with_transaction=False)
        with self.assertRaises(exceptions.IntegrityError):
            with client.connect() as conn:
                # Make some change in a table.
                conn.execute("""
                INSERT INTO records
                VALUES ('rock-and-roll', 'music', 'genre', NOW(), '{}', FALSE);
                """)
                # Go into a failing integrity constraint.
                query = "INSERT INTO timestamps VALUES ('a', 'b', NOW());"
                conn.execute(query)
                conn.execute(query)
                conn.commit()
                conn.close()

        # Check that change in the above table was rolledback.
        with client.connect() as conn:
            result = conn.execute("""
            SELECT FROM records
             WHERE parent_id = 'music'
               AND collection_id = 'genre';
            """)
        self.assertEqual(result.rowcount, 0)

    def test_conflicts_handled_correctly(self):
        config = self._get_config()
        storage = self.backend.load_from_config(config)
        storage.create(collection_id='genre', parent_id='music',
                       record={'id': 'rock-and-roll'})

        def record_not_found(*args, **kwargs):
            raise exceptions.RecordNotFoundError()

        with mock.patch.object(storage, 'get', side_effect=record_not_found):
            with self.assertRaises(exceptions.UnicityError):
                storage.create(collection_id='genre', parent_id='music',
                               record={'id': 'rock-and-roll'})


class PaginatedTest(unittest.TestCase):
    def setUp(self):
        self.storage = mock.Mock()
        self.sample_records = [
            {"id": "record-01", "flavor": "strawberry"},
            {"id": "record-02", "flavor": "banana"},
            {"id": "record-03", "flavor": "mint"},
            {"id": "record-04", "flavor": "plain"},
            {"id": "record-05", "flavor": "peanut"},
        ]

        def sample_records_side_effect(*args, **kwargs):
            return (self.sample_records, len(self.sample_records))

        self.storage.get_all.side_effect = sample_records_side_effect

    def test_paginated_passes_sort(self):
        i = paginated(self.storage, sorting=[Sort('id', -1)])
        next(i)   # make the generator do anything
        self.storage.get_all.assert_called_with(sorting=[Sort('id', -1)],
                                                limit=25, pagination_rules=None)

    def test_paginated_passes_batch_size(self):
        i = paginated(self.storage, sorting=[Sort('id', -1)], batch_size=17)
        next(i)   # make the generator do anything
        self.storage.get_all.assert_called_with(sorting=[Sort('id', -1)],
                                                limit=17, pagination_rules=None)

    def test_paginated_yields_records(self):
        iter = paginated(self.storage, sorting=[Sort('id', -1)])
        assert next(iter) == {"id": "record-01", "flavor": "strawberry"}

    def test_paginated_fetches_next_page(self):
        records = self.sample_records
        records.reverse()

        def get_all_mock(*args, **kwargs):
            this_records = records[:3]
            del records[:3]
            return this_records, len(this_records)

        self.storage.get_all.side_effect = get_all_mock

        list(paginated(self.storage, sorting=[Sort('id', -1)]))
        assert self.storage.get_all.call_args_list == [
            mock.call(sorting=[Sort('id', -1)], limit=25, pagination_rules=None),
            mock.call(sorting=[Sort('id', -1)], limit=25, pagination_rules=[
                [Filter('id', 'record-03', COMPARISON.LT)]
            ]),
            mock.call(sorting=[Sort('id', -1)], limit=25, pagination_rules=[
                [Filter('id', 'record-01', COMPARISON.LT)]
            ]),
        ]



import urllib.parse

import colander
import requests
from pyramid import httpexceptions

from cornice.validators import colander_validator
from kinto.core import Service
from kinto.core.errors import raise_invalid, ERRORS
from kinto.core.utils import random_bytes_hex
from kinto.core.resource.schema import ErrorResponseSchema
from kinto.core.schema import URL

from .utils import fetch_openid_config


DEFAULT_STATE_TTL_SECONDS = 3600
DEFAULT_STATE_LENGTH = 32


class RedirectHeadersSchema(colander.MappingSchema):
    """Redirect response headers."""
    location = colander.SchemaNode(colander.String(), name='Location')


class RedirectResponseSchema(colander.MappingSchema):
    """Redirect response schema."""
    headers = RedirectHeadersSchema()


response_schemas = {
    '307': RedirectResponseSchema(description='Successful redirection.'),
    '400': ErrorResponseSchema(description='The request is invalid.'),
}


def provider_validator(request, **kwargs):
    """
    This validator verifies that the validator in URL (eg. /openid/auth0/login)
    is a configured OpenIDConnect policy.
    """
    provider = request.matchdict['provider']
    used = request.registry.settings.get('multiauth.policy.%s.use' % provider, '')
    if not used.endswith('OpenIDConnectPolicy'):
        request.errors.add('path', 'provider', 'Unknow provider %r' % provider)


class LoginQuerystringSchema(colander.MappingSchema):
    """
    Querystring schema for the login endpoint.
    """
    callback = URL()
    scope = colander.SchemaNode(colander.String())
    prompt = colander.SchemaNode(colander.String(),
                                 validator=colander.Regex("none"),
                                 missing=colander.drop)


class LoginSchema(colander.MappingSchema):
    querystring = LoginQuerystringSchema()


login = Service(name='openid_login',
                path='/openid/{provider}/login',
                description='Initiate the OAuth2 login')


@login.get(schema=LoginSchema(),
           validators=(colander_validator, provider_validator),
           response_schemas=response_schemas)
def get_login(request):
    """Initiates to login dance for the specified scopes and callback URI
    using appropriate redirections."""

    # Settings.
    provider = request.matchdict['provider']
    settings_prefix = 'multiauth.policy.%s.' % provider
    issuer = request.registry.settings[settings_prefix + 'issuer']
    client_id = request.registry.settings[settings_prefix + 'client_id']
    userid_field = request.registry.settings.get(settings_prefix + 'userid_field')
    state_ttl = int(request.registry.settings.get(settings_prefix + 'state_ttl_seconds',
                                                  DEFAULT_STATE_TTL_SECONDS))
    state_length = int(request.registry.settings.get(settings_prefix + 'state_length',
                                                     DEFAULT_STATE_LENGTH))

    # Read OpenID configuration (cached by issuer)
    oid_config = fetch_openid_config(issuer)
    auth_endpoint = oid_config['authorization_endpoint']

    scope = request.GET['scope']
    callback = request.GET['callback']
    prompt = request.GET.get('prompt')

    # Check that email scope is requested if userid field is configured as email.
    if userid_field == 'email' and 'email' not in scope:
        error_details = {
            'name': 'scope',
            'description': "Provider %s requires 'email' scope" % provider,
        }
        raise_invalid(request, **error_details)

    # Generate a random string as state.
    # And save it until code is traded.
    state = random_bytes_hex(state_length)
    request.registry.cache.set('openid:state:' + state, callback, ttl=state_ttl)

    # Redirect the client to the Identity Provider that will eventually redirect
    # to the OpenID token endpoint.
    token_uri = request.route_url('openid_token', provider=provider) + '?'
    params = dict(client_id=client_id, response_type='code', scope=scope,
                  redirect_uri=token_uri, state=state)
    if prompt:
        # The 'prompt' parameter is optional.
        params['prompt'] = prompt
    redirect = '{}?{}'.format(auth_endpoint, urllib.parse.urlencode(params))
    raise httpexceptions.HTTPTemporaryRedirect(redirect)


class TokenQuerystringSchema(colander.MappingSchema):
    """
    Querystring schema for the token endpoint.
    """
    code = colander.SchemaNode(colander.String())
    state = colander.SchemaNode(colander.String())


class TokenSchema(colander.MappingSchema):
    querystring = TokenQuerystringSchema()


token = Service(name='openid_token',
                path='/openid/{provider}/token',
                description='')


@token.get(schema=TokenSchema(),
           validators=(colander_validator, provider_validator))
def get_token(request):
    """Trades the specified code and state against access and ID tokens.
    The client is redirected to the original ``callback`` URI with the
    result in querystring."""

    # Settings.
    provider = request.matchdict['provider']
    settings_prefix = 'multiauth.policy.%s.' % provider
    issuer = request.registry.settings[settings_prefix + 'issuer']
    client_id = request.registry.settings[settings_prefix + 'client_id']
    client_secret = request.registry.settings[settings_prefix + 'client_secret']

    # Read OpenID configuration (cached by issuer)
    oid_config = fetch_openid_config(issuer)
    token_endpoint = oid_config['token_endpoint']

    code = request.GET['code']
    state = request.GET['state']

    # State can be used only once.
    callback = request.registry.cache.delete('openid:state:' + state)
    if callback is None:
        error_details = {
            'name': 'state',
            'description': 'Invalid state',
            'errno': ERRORS.INVALID_AUTH_TOKEN.value,
        }
        raise_invalid(request, **error_details)

    # Trade the code for tokens on the Identity Provider.
    # Google Identity requires to specify again redirect_uri.
    redirect_uri = request.route_url('openid_token', provider=provider) + '?'
    data = {
        'code': code,
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code',
    }
    resp = requests.post(token_endpoint, data=data)

    # The IdP response is forwarded to the client in the querystring/location hash.
    # (eg. callback=`http://localhost:3000/#tokens=`)
    redirect = callback + urllib.parse.quote(resp.text)
    raise httpexceptions.HTTPTemporaryRedirect(redirect)



import re

import requests
from pyramid import authentication as base_auth
from pyramid.interfaces import IAuthenticationPolicy
from zope.interface import implementer

from kinto.core import logger
from kinto.core import utils as core_utils
from kinto.core.openapi import OpenAPI

from .utils import fetch_openid_config


@implementer(IAuthenticationPolicy)
class OpenIDConnectPolicy(base_auth.CallbackAuthenticationPolicy):
    def __init__(self, issuer, client_id, realm='Realm', **kwargs):
        self.realm = realm
        self.issuer = issuer
        self.client_id = client_id
        self.client_secret = kwargs.get('client_secret', '')
        self.header_type = kwargs.get('header_type', 'Bearer')
        self.userid_field = kwargs.get('userid_field', 'sub')
        self.verification_ttl = int(kwargs.get('verification_ttl_seconds', 86400))

        # Fetch OpenID config (at instantiation, ie. startup)
        self.oid_config = fetch_openid_config(issuer)

        self._jwt_keys = None

    def unauthenticated_userid(self, request):
        """Return the userid or ``None`` if token could not be verified.
        """
        settings = request.registry.settings
        hmac_secret = settings['userid_hmac_secret']

        authorization = request.headers.get('Authorization', '')
        try:
            authmeth, access_token = authorization.split(' ', 1)
        except ValueError:
            return None

        if authmeth.lower() != self.header_type.lower():
            return None

        # XXX JWT Access token
        # https://auth0.com/docs/tokens/access-token#access-token-format

        # Check cache if these tokens were already verified.
        hmac_tokens = core_utils.hmac_digest(hmac_secret, access_token)
        cache_key = 'openid:verify:{}'.format(hmac_tokens)
        payload = request.registry.cache.get(cache_key)
        if payload is None:
            # This can take some time.
            payload = self._verify_token(access_token)
            if payload is None:
                return None
        # Save for next time / refresh ttl.
        request.registry.cache.set(cache_key, payload, ttl=self.verification_ttl)
        # Extract meaningful field from userinfo (eg. email or sub)
        return payload.get(self.userid_field)

    def forget(self, request):
        """A no-op. Credentials are sent on every request.
        Return WWW-Authenticate Realm header for Bearer token.
        """
        return [('WWW-Authenticate', '%s realm="%s"' % (self.header_type, self.realm))]

    def _verify_token(self, access_token):
        uri = self.oid_config['userinfo_endpoint']
        # Opaque access token string. Fetch user info from profile.
        try:
            resp = requests.get(uri, headers={'Authorization': 'Bearer ' + access_token})
            resp.raise_for_status()
            userprofile = resp.json()
            return userprofile

        except (requests.exceptions.HTTPError, ValueError, KeyError) as e:
            logger.debug('Unable to fetch user profile from %s (%s)' % (uri, e))
            return None


def includeme(config):
    # Activate end-points.
    config.scan('kinto.plugins.openid.views')

    settings = config.get_settings()

    openid_policies = []
    for k, v in settings.items():
        m = re.match('multiauth\.policy\.(.*)\.use', k)
        if m:
            if v.endswith('OpenIDConnectPolicy'):
                openid_policies.append(m.group(1))

    if len(openid_policies) == 0:
        # Do not add the capability if no policy is configured.
        return

    providers_infos = []
    for name in openid_policies:
        issuer = settings['multiauth.policy.%s.issuer' % name]
        openid_config = fetch_openid_config(issuer)

        client_id = settings['multiauth.policy.%s.client_id' % name]
        header_type = settings.get('multiauth.policy.%s.header_type', 'Bearer')

        providers_infos.append({
            'name': name,
            'issuer': openid_config['issuer'],
            'auth_path': '/openid/%s/login' % name,
            'client_id': client_id,
            'header_type': header_type,
            'userinfo_endpoint': openid_config['userinfo_endpoint'],
        })

        OpenAPI.expose_authentication_method(name, {
            'type': 'oauth2',
            'authorizationUrl': openid_config['authorization_endpoint'],
        })

    config.add_api_capability(
        'openid',
        description='OpenID connect support.',
        url='http://kinto.readthedocs.io/en/stable/api/1.x/authentication.html',
        providers=providers_infos)



import unittest
import mock

from kinto.core.testing import DummyRequest
from kinto.plugins.openid import OpenIDConnectPolicy
from kinto.plugins.openid.utils import fetch_openid_config

from .. import support


def get_openid_configuration(url):
    base_url = url.replace('/.well-known/openid-configuration', '')
    m = mock.Mock()
    m.json.return_value = {
        'issuer': '{base_url} issuer'.format(base_url=base_url),
        'authorization_endpoint': '{base_url}/authorize'.format(base_url=base_url),
        'userinfo_endpoint': '{base_url}/oauth/user'.format(base_url=base_url),
        'token_endpoint': '{base_url}/oauth/token'.format(base_url=base_url),
    }
    return m


class OpenIDWebTest(support.BaseWebTest, unittest.TestCase):

    @classmethod
    def make_app(cls, *args, **kwargs):
        with mock.patch('kinto.plugins.openid.requests.get') as get:
            get.side_effect = get_openid_configuration
            return super(OpenIDWebTest, cls).make_app(*args, **kwargs)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        openid_policy = 'kinto.plugins.openid.OpenIDConnectPolicy'
        settings['includes'] = 'kinto.plugins.openid'
        settings['multiauth.policies'] = 'auth0 google'
        settings['multiauth.policy.auth0.use'] = openid_policy
        settings['multiauth.policy.auth0.issuer'] = 'https://auth.mozilla.auth0.com'
        settings['multiauth.policy.auth0.client_id'] = 'abc'
        settings["multiauth.policy.auth0.client_secret"] = 'xyz'

        settings['multiauth.policy.google.use'] = openid_policy
        settings['multiauth.policy.google.issuer'] = 'https://accounts.google.com'
        settings['multiauth.policy.google.client_id'] = '123'
        settings["multiauth.policy.google.client_secret"] = '789'
        settings["multiauth.policy.google.userid_field"] = 'email'
        return settings


class OpenIDWithoutPolicyTest(support.BaseWebTest, unittest.TestCase):

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = 'kinto.plugins.openid'
        return settings

    def test_openid_capability_is_not_added(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        assert 'openid' not in capabilities


class HelloViewTest(OpenIDWebTest):

    def test_openid_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        assert 'openid' in capabilities
        assert len(capabilities['openid']['providers']) == 2
        assert 'userinfo_endpoint' in capabilities['openid']['providers'][0]
        assert 'auth_path' in capabilities['openid']['providers'][0]

    def test_openid_in_openapi(self):
        resp = self.app.get('/__api__')
        assert 'auth0' in resp.json['securityDefinitions']
        auth = resp.json['securityDefinitions']['auth0']['authorizationUrl']
        assert auth == 'https://auth.mozilla.auth0.com/authorize'


class PolicyTest(unittest.TestCase):
    def setUp(self):
        mocked = mock.patch('kinto.plugins.openid.requests.get')
        self.mocked_get = mocked.start()
        self.addCleanup(mocked.stop)

        self.policy = OpenIDConnectPolicy(issuer='https://idp', client_id='abc')

        self.request = DummyRequest()
        self.request.registry.cache.get.return_value = None

        mocked = mock.patch.object(self.policy, '_verify_token')
        self.verify = mocked.start()
        self.addCleanup(mocked.stop)
        self.verify.return_value = {'sub': 'userid'}

    def test_returns_none_if_no_authorization(self):
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_returns_header_type_in_forget(self):
        h = self.policy.forget(self.request)
        assert 'Bearer ' in h[0][1]

    def test_header_type_can_be_configured(self):
        self.policy.header_type = 'bearer+oidc'
        h = self.policy.forget(self.request)
        assert 'bearer+oidc ' in h[0][1]

    def test_returns_none_if_no_authorization_prefix(self):
        self.request.headers['Authorization'] = 'avrbnnbrbr'
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_returns_none_if_bad_prefix(self):
        self.request.headers['Authorization'] = 'Basic avrbnnbrbr'
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_can_specify_only_opaque_access_token(self):
        self.request.headers['Authorization'] = 'Bearer xyz'
        assert self.policy.unauthenticated_userid(self.request) == 'userid'
        self.verify.assert_called_with('xyz')

    def test_returns_none_if_no_cache_and_invalid_access_token(self):
        self.request.headers['Authorization'] = 'Bearer xyz'
        self.request.registry.cache.get.return_value = None
        self.verify.return_value = None
        assert self.policy.unauthenticated_userid(self.request) is None
        assert not self.request.registry.cache.set.called

    def test_payload_is_read_from_cache(self):
        self.request.headers['Authorization'] = 'Bearer xyz'
        self.request.registry.cache.get.return_value = {'sub': 'me'}
        assert self.policy.unauthenticated_userid(self.request) == 'me'

    def test_payload_is_stored_in_cache(self):
        self.request.headers['Authorization'] = 'Bearer xyz'
        assert self.policy.unauthenticated_userid(self.request) == 'userid'
        assert self.request.registry.cache.set.called

    def test_payload_is_read_from_cache_but_differently_by_access_token(self):
        # State to keep track of cache keys queried.
        cache_keys_used = []

        def mocked_cache_get(cache_key):
            # This makes sure the same cache key is not used twice
            assert cache_key not in cache_keys_used
            cache_keys_used.append(cache_key)
            if len(cache_keys_used) == 1:
                return {'sub': 'me'}
            elif len(cache_keys_used) == 2:
                return {'sub': 'you'}

        self.request.registry.cache.get.side_effect = mocked_cache_get

        self.request.headers['Authorization'] = 'Bearer xyz'
        assert self.policy.unauthenticated_userid(self.request) == 'me'

        # Change the Authorization header the second time
        self.request.headers['Authorization'] = 'Bearer abc'
        assert self.policy.unauthenticated_userid(self.request) == 'you'


class VerifyTokenTest(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        # Populate OpenID config cache.
        with mock.patch('kinto.plugins.openid.utils.requests.get') as m:
            m.return_value.json.return_value = {'userinfo_endpoint': 'http://uinfo',
                                                'jwks_uri': 'https://jwks'}
            fetch_openid_config('https://fxa')

    def setUp(self):
        mocked = mock.patch('kinto.plugins.openid.requests.get')
        self.mocked_get = mocked.start()
        self.addCleanup(mocked.stop)

        self.policy = OpenIDConnectPolicy(issuer='https://fxa', client_id='abc')

    def test_fetches_userinfo_if_id_token_is_none(self):
        self.mocked_get.return_value.json.side_effect = [
            {"sub": "me"},
        ]
        payload = self.policy._verify_token(access_token='abc')
        assert payload["sub"] == "me"

    def test_returns_none_if_fetching_userinfo_fails(self):
        self.mocked_get.return_value.raise_for_status.side_effect = ValueError
        payload = self.policy._verify_token(access_token='abc')
        assert payload is None


class LoginViewTest(OpenIDWebTest):

    def test_returns_400_if_parameters_are_missing_or_bad(self):
        self.app.get('/openid/auth0/login', status=400)
        self.app.get('/openid/auth0/login', params={'callback': 'http://no-scope'}, status=400)
        self.app.get('/openid/auth0/login', params={'callback': 'bad', 'scope': 'openid'},
                     status=400)

    def test_returns_400_if_provider_is_unknown(self):
        self.app.get('/openid/fxa/login', status=400)

    def test_returns_400_if_email_is_not_in_scope_when_userid_field_is_email(self):
        scope = 'openid'
        cb = 'http://ui'
        self.app.get('/openid/auth0/login', params={'callback': cb, 'scope': scope},
                     status=307)
        # See config above (email is userid field)
        self.app.get('/openid/google/login', params={'callback': cb, 'scope': scope},
                     status=400)

    def test_returns_400_if_prompt_is_not_recognized(self):
        scope = 'openid'
        cb = 'http://ui'
        self.app.get('/openid/auth0/login',
                     params={'callback': cb, 'scope': scope, 'prompt': 'junk'},
                     status=400)

    def test_redirects_to_the_identity_provider(self):
        params = {'callback': 'http://ui', 'scope': 'openid'}
        resp = self.app.get('/openid/auth0/login', params=params, status=307)
        location = resp.headers['Location']
        assert 'auth0.com/authorize?' in location
        assert '%2Fv1%2Fopenid%2Fauth0%2Ftoken' in location
        assert 'scope=openid' in location
        assert 'client_id=abc' in location

    def test_redirects_to_the_identity_provider_with_prompt_none(self):
        params = {'callback': 'http://ui', 'scope': 'openid', 'prompt': 'none'}
        resp = self.app.get('/openid/auth0/login', params=params, status=307)
        location = resp.headers['Location']
        assert 'auth0.com/authorize?' in location
        assert '%2Fv1%2Fopenid%2Fauth0%2Ftoken' in location
        assert 'scope=openid' in location
        assert 'client_id=abc' in location
        assert 'prompt=none' in location

    def test_callback_is_stored_in_cache(self):
        params = {'callback': 'http://ui', 'scope': 'openid'}
        with mock.patch('kinto.plugins.openid.views.random_bytes_hex') as m:
            m.return_value = 'key'
            self.app.get('/openid/auth0/login', params=params, status=307)

        cached = self.app.app.registry.cache.get('openid:state:key')
        assert cached == 'http://ui'


class TokenViewTest(OpenIDWebTest):

    def test_returns_400_if_parameters_are_missing_or_bad(self):
        self.app.get('/openid/auth0/token', status=400)
        self.app.get('/openid/auth0/token', params={'code': 'abc'}, status=400)
        self.app.get('/openid/auth0/token', params={'state': 'abc'}, status=400)

    def test_returns_400_if_provider_is_unknown(self):
        self.app.get('/openid/fxa/token', status=400)

    def test_returns_400_if_state_is_invalid(self):
        self.app.get('/openid/auth0/token', params={'code': 'abc', 'state': 'abc'}, status=400)

    def test_code_is_traded_using_client_secret(self):
        self.app.app.registry.cache.set('openid:state:key', 'http://ui', ttl=100)
        with mock.patch('kinto.plugins.openid.views.requests.post') as m:
            m.return_value.text = '{"access_token": "token"}'
            self.app.get('/openid/auth0/token', params={'code': 'abc', 'state': 'key'})
            m.assert_called_with(
                'https://auth.mozilla.auth0.com/oauth/token',
                data={
                    'code': 'abc',
                    'client_id': 'abc',
                    'client_secret': 'xyz',
                    'redirect_uri': 'http://localhost/v1/openid/auth0/token?',
                    'grant_type': 'authorization_code'})

    def test_state_cannot_be_reused(self):
        self.app.app.registry.cache.set('openid:state:key', 'http://ui', ttl=100)
        with mock.patch('kinto.plugins.openid.views.requests.post') as m:
            m.return_value.text = '{"access_token": "token"}'
            self.app.get('/openid/auth0/token', params={'code': 'abc', 'state': 'key'})
            self.app.get('/openid/auth0/token', params={'code': 'abc', 'state': 'key'}, status=400)

    def test_redirects_to_callback_using_authorization_response(self):
        self.app.app.registry.cache.set('openid:state:key', 'http://ui/#token=', ttl=100)
        with mock.patch('kinto.plugins.openid.views.requests.post') as m:
            m.return_value.text = '{"access_token": "token"}'
            resp = self.app.get('/openid/auth0/token', params={'code': 'abc', 'state': 'key'},
                                status=307)
        location = resp.headers['Location']
        assert location == 'http://ui/#token=%7B%22access_token%22%3A%20%22token%22%7D'



import json
import os

from pyramid.static import static_view
from pyramid.httpexceptions import HTTPTemporaryRedirect

from .views import admin_home_view


HERE = os.path.dirname(__file__)


def includeme(config):
    # Process settings to remove storage wording.

    # Read version from package.json
    package_json = json.load(open(os.path.join(HERE, 'package.json')))
    admin_version = package_json['dependencies']['kinto-admin']

    # Expose capability.
    config.add_api_capability(
        'admin',
        version=admin_version,
        description='Serves the admin console.',
        url='https://github.com/Kinto/kinto-admin/',
    )

    config.add_route('admin_home', '/admin/')
    config.add_view(admin_home_view, route_name='admin_home')

    build_dir = static_view('kinto.plugins.admin:build', use_subpath=True)
    config.add_route('catchall_static', '/admin/*subpath')
    config.add_view(build_dir, route_name='catchall_static')

    # Setup redirect without trailing slash.
    def admin_redirect_view(request):
        raise HTTPTemporaryRedirect(request.path + '/')

    config.add_route('admin_redirect', '/admin')
    config.add_view(admin_redirect_view, route_name='admin_redirect')



import random
from base64 import b64encode, b64decode

import mock
from urllib.parse import parse_qs, urlparse
from pyramid.httpexceptions import HTTPBadRequest

from kinto.core.utils import json

from . import BaseTest


class BasePaginationTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()

        indices = list(range(20))
        random.shuffle(indices)
        for i in indices:
            record = {
                'title': 'MoFo #{0:02}'.format(i),
                'status': i % 4,
                'unread': (i % 2 == 0)
            }
            if i % 3 == 0:
                record['optional'] = True
            self.model.create_record(record)

        self.validated = self.resource.request.validated

    def _setup_next_page(self):
        next_page = self.last_response.headers['Next-Page']
        url_fragments = urlparse(next_page)
        queryparams = parse_qs(url_fragments.query)
        self.validated['querystring']['_token'] = queryparams['_token'][0]
        self.validated['querystring']['_limit'] = int(queryparams['_limit'][0])
        self.last_response.headers = {}
        return queryparams


class PaginationTest(BasePaginationTest):
    def test_return_data(self):
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 20)

    def test_handle_limit(self):
        self.validated['querystring'] = {'_limit': 10}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 10)

    def test_handle_forced_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [
                ('paginate_by', 10)]):
            result = self.resource.collection_get()
            self.assertEqual(len(result['data']), 10)

    def test_forced_limit_has_precedence_over_provided_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [
                ('paginate_by', 5)]):
            self.validated['querystring'] = {'_limit': 10}
            result = self.resource.collection_get()
            self.assertEqual(len(result['data']), 5)

    def test_return_total_records_in_headers(self):
        self.validated['querystring'] = {'_limit': 5}
        self.resource.collection_get()
        headers = self.last_response.headers
        count = headers['Total-Records']
        self.assertEquals(int(count), 20)

    def test_return_next_page_url_is_given_in_headers(self):
        self.validated['querystring'] = {'_limit': 10}
        self.resource.collection_get()
        self.assertIn('Next-Page', self.last_response.headers)

    def test_next_page_url_has_got_querystring(self):
        self.validated['querystring'] = {'_limit': 10}
        self.resource.collection_get()
        queryparams = self._setup_next_page()
        self.assertIn('_limit', queryparams)
        self.assertIn('_token', queryparams)

    def test_next_page_url_gives_distinct_records(self):
        self.validated['querystring'] = {'_limit': 10}
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        results_id1 = set([x['id'] for x in results1['data']])
        results_id2 = set([x['id'] for x in results2['data']])
        self.assertFalse(results_id1.intersection(results_id2))

    def test_next_page_url_gives_distinct_records_with_forced_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [
                ('paginate_by', 5)]):
            results1 = self.resource.collection_get()
            self._setup_next_page()
            results2 = self.resource.collection_get()

            results_id1 = set([x['id'] for x in results1['data']])
            results_id2 = set([x['id'] for x in results2['data']])
            self.assertFalse(results_id1.intersection(results_id2))

    def test_twice_the_same_next_page(self):
        self.validated['querystring'] = {'_limit': 10}
        records = self.resource.collection_get()
        first_ids = [r['id'] for r in records['data']]
        records = self.resource.collection_get()
        second_ids = [r['id'] for r in records['data']]
        self.assertEqual(first_ids, second_ids)

    def test_stops_giving_next_page_at_the_end_of_first_page(self):
        self.resource.collection_get()
        self.assertNotIn('Next-Page', self.last_response.headers)

    def test_stops_giving_next_page_at_the_end_sets(self):
        self.validated['querystring'] = {'_limit': 11}
        self.resource.collection_get()
        self._setup_next_page()
        self.resource.collection_get()
        self.assertNotIn('Next-Page', self.last_response.headers)

    def test_stops_giving_next_page_at_the_end_sets_on_exact_limit(self):
        self.validated['querystring'] = {'_limit': 10}
        self.resource.collection_get()
        self._setup_next_page()
        self.resource.collection_get()
        self.assertNotIn('Next-Page', self.last_response.headers)

    def test_handle_simple_sorting(self):
        self.validated['querystring'] = {'_sort': ['-status'], '_limit': 20}
        expected_results = self.resource.collection_get()
        self.validated['querystring']['_limit'] = 10
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        self.assertEqual(expected_results['data'],
                         results1['data'] + results2['data'])

    def test_handle_multiple_sorting(self):
        self.validated['querystring'] = {'_sort': ['-status', 'title'], '_limit': 20}
        expected_results = self.resource.collection_get()
        self.validated['querystring']['_limit'] = 10
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        self.assertEqual(expected_results['data'],
                         results1['data'] + results2['data'])

    def test_handle_filtering_sorting(self):
        self.validated['querystring'] = {'_sort': ['-status', 'title'], 'status': 2,
                                         '_limit': 20}
        expected_results = self.resource.collection_get()
        self.validated['querystring']['_limit'] = 3
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        self.assertEqual(expected_results['data'],
                         results1['data'] + results2['data'])

    def test_handle_sorting_desc(self):
        self.validated['querystring'] = {'_sort': ['status', '-title'], '_limit': 20}
        expected_results = self.resource.collection_get()
        self.validated['querystring']['_limit'] = 10
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        self.assertEqual(expected_results['data'],
                         results1['data'] + results2['data'])

    def test_handle_since(self):
        self.validated['querystring'] = {'_since': 123, '_limit': 20}
        expected_results = self.resource.collection_get()
        self.validated['querystring']['_limit'] = 10
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        self.assertEqual(expected_results['data'],
                         results1['data'] + results2['data'])

    def test_token_wrong_base64(self):
        self.validated['querystring'] = {
            '_since': 123, '_limit': 20, '_token': '123'}
        self.assertRaises(HTTPBadRequest, self.resource.collection_get)

    def test_token_wrong_json(self):
        self.validated['querystring'] = {
            '_since': 123, '_limit': 20,
            '_token': b64encode('{"toto":'.encode('ascii')).decode('ascii')}
        self.assertRaises(HTTPBadRequest, self.resource.collection_get)

    def test_token_wrong_json_fields(self):
        badtoken = '{"toto": {"tutu": 1}}'
        self.validated['querystring'] = {
            '_since': 123, '_limit': 20,
            '_token': b64encode(badtoken.encode('ascii')).decode('ascii')}
        self.assertRaises(HTTPBadRequest, self.resource.collection_get)

    def test_raises_bad_request_if_token_has_bad_data_structure(self):
        invalid_token = json.dumps([[('last_modified', 0, '>')]])
        self.validated['querystring'] = {
            '_since': 123, '_limit': 20,
            '_token': b64encode(invalid_token.encode('ascii')).decode('ascii')}
        self.assertRaises(HTTPBadRequest, self.resource.collection_get)

    def test_next_page_url_works_with_optional_fields(self):
        self.validated['querystring'] = {'_limit': 10, '_sort': ['-optional']}
        results1 = self.resource.collection_get()
        self._setup_next_page()
        results2 = self.resource.collection_get()
        results_id1 = set([x['id'] for x in results1['data']])
        results_id2 = set([x['id'] for x in results2['data']])
        self.assertFalse(results_id1.intersection(results_id2))


class PaginatedDeleteTest(BasePaginationTest):
    def test_handle_limit_on_delete(self):
        self.validated['querystring'] = {'_limit': 3}
        result = self.resource.collection_delete()
        self.assertEqual(len(result['data']), 3)

    def test_paginated_delete(self):
        all_records = self.resource.collection_get()
        expected_ids = [r['id'] for r in all_records['data']]
        # Page 1
        self.validated['querystring']['_limit'] = 10
        results1 = self.resource.collection_delete()
        results1_ids = [r['id'] for r in results1['data']]
        self._setup_next_page()
        # Page 2
        results2 = self.resource.collection_delete()
        results2_ids = [r['id'] for r in results2['data']]
        self.assertEqual(expected_ids, results1_ids + results2_ids)

    def test_return_total_records_in_headers_matching_deletable(self):
        self.validated['querystring'] = {'_limit': 5}
        self.resource.collection_delete()
        headers = self.last_response.headers
        count = headers['Total-Records']
        self.assertEquals(int(count), 20)

    def test_paginated_delete_second_to_last_gets_next_header(self):
        self.resource.collection_get()
        get_all_headers = self.last_response.headers
        count = int(get_all_headers['Total-Records']) - 1

        self.validated['querystring'] = {'_limit': 1}
        headers = []
        for i in range(count):
            self.resource.collection_delete()
            headers.append(self.last_response.headers)
            self._setup_next_page()

        self.resource.collection_delete()
        headers.append(self.last_response.headers)

        self.assertIn('Next-Page', headers[count - 1])
        self.assertNotIn('Next-Page', headers[count])

    def test_token_cannot_be_reused_twice(self):
        self.resource.request.method = "DELETE"
        self.validated['querystring']['_limit'] = 3
        self.resource.collection_delete()
        self._setup_next_page()
        self.resource.collection_delete()
        # Reuse previous token.
        self.assertRaises(HTTPBadRequest, self.resource.collection_delete)


class BuildPaginationTokenTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()
        self.record = {
            'id': 1, 'status': 2, 'unread': True,
            'last_modified': 1234, 'title': 'Title',
            'nested': {'subvalue': 42},
            'nested.other': {'subvalue': 43},
        }

    def test_token_contains_current_offset(self):
        token = self.resource._build_pagination_token([('last_modified', -1)],
                                                      self.record,
                                                      42)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['offset'], 42)

    def test_no_sorting_default_to_modified_field(self):
        token = self.resource._build_pagination_token([('last_modified', -1)],
                                                      self.record,
                                                      42)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertDictEqual(tokeninfo['last_record'],
                             {"last_modified": 1234})

    def test_sorting_handle_both_rules(self):
        token = self.resource._build_pagination_token([
            ('status', -1),
            ('last_modified', -1)
        ], self.record, 34)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertDictEqual(tokeninfo['last_record'],
                             {"last_modified": 1234, "status": 2})

    def test_sorting_handle_ordering_direction(self):
        token = self.resource._build_pagination_token([
            ('status', 1),
            ('last_modified', 1)
        ], self.record, 32)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {"last_modified": 1234, "status": 2})

    def test_multiple_sorting_keep_all(self):
        token = self.resource._build_pagination_token([
            ('status', 1),
            ('title', -1),
            ('last_modified', -1)
        ], self.record, 31)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {"last_modified": 1234, "status": 2,
                          'title': 'Title'})

    def test_sorting_on_nested_field(self):
        token = self.resource._build_pagination_token([
            ('nested.subvalue', -1),
            ('title', 1),
        ], self.record, 88)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {'title': 'Title', 'nested.subvalue': 42})

    def test_disambiguate_fieldname_containing_dots(self):
        token = self.resource._build_pagination_token([
            ('nested.other.subvalue', -1),
            ('title', 1),
        ], self.record, 88)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {'title': 'Title', 'nested.other.subvalue': 43})

    def test_strip_malformed_sort_field(self):
        token = self.resource._build_pagination_token([
            ('non.existent', -1),
            ('title', 1),
        ], self.record, 88)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {'title': 'Title'})

    def test_can_build_while_sorting_on_missing_field(self):
        token = self.resource._build_pagination_token([
            ('unknown', 1),
            ('title', -1),
            ('last_modified', -1)
        ], self.record, 31)
        tokeninfo = json.loads(b64decode(token).decode('ascii'))
        self.assertEqual(tokeninfo['last_record'],
                         {"last_modified": 1234, "title": "Title"})



import colander
from pyramid.security import NO_PERMISSION_REQUIRED
from cornice.service import get_services

from kinto.core import Service
from kinto.core.openapi import OpenAPI


openapi = Service(name='openapi', path='/__api__', description='OpenAPI description')


class OpenAPIResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


openapi_response_schemas = {
    '200': OpenAPIResponseSchema(
        description='Return an OpenAPI description of the running instance.')
}


@openapi.get(permission=NO_PERMISSION_REQUIRED,
             response_schemas=openapi_response_schemas,
             tags=['Utilities'], operation_id='get_openapi_spec')
def openapi_view(request):

    # Only build json once
    try:
        return openapi_view.__json__
    except AttributeError:
        openapi_view.__json__ = OpenAPI(get_services(), request).generate()
        return openapi_view.__json__



import mock
import uuid
from contextlib import contextmanager

from pyramid.config import Configurator

from kinto.core import statsd
from kinto.core.events import (ResourceChanged, AfterResourceChanged,
                               ResourceRead, AfterResourceRead, ACTIONS,
                               notify_resource_event)
from kinto.core.storage.exceptions import BackendError
from kinto.core.testing import unittest

from ..support import BaseWebTest


@contextmanager
def notif_broken(app, event_cls):
    old = app.registry.notify

    def buggy(event):
        if not isinstance(event, event_cls):
            return old(event)
        raise Exception("boom")

    app.registry.notify = buggy
    yield
    app.registry.notify = old


class BaseEventTest(BaseWebTest):

    subscribed = tuple()
    events = []

    def setUp(self):
        super().setUp()
        del self.events[:]
        self.body = {'data': {'name': 'de Paris'}}

    def tearDown(self):
        del self.events[:]
        super().tearDown()

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

    @classmethod
    def make_app(cls, settings=None, config=None):
        settings = cls.get_app_settings(settings)
        config = Configurator(settings=settings)
        for event_cls in cls.subscribed:
            config.add_subscriber(cls.listener, event_cls)
        config.commit()
        return super().make_app(settings=settings, config=config)


class ResourceReadTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceRead,)

    def test_get_sends_read_event(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record_id = resp.json['data']['id']
        record_url = self.get_item_url(record_id)
        self.app.get(record_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_records), 1)

    def test_collection_get_sends_read_event(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_records), 0)

    def test_post_sends_read_if_id_already_exists(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        body = {**self.body}
        body['data']['id'] = record['id']

        # a second post with the same record id
        self.app.post_json(self.collection_url, body, headers=self.headers,
                           status=200)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)


class ResourceChangedTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged,)

    def test_events_have_custom_representation(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.assertEqual(repr(self.events[0]),
                         "<ResourceChanged action=create "
                         "uri={}>".format(self.collection_url))

    def test_post_sends_create_action(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_put_sends_create_action(self):
        body = {**self.body}
        body['data']['id'] = record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, body,
                          headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_not_triggered_on_failed_put(self):
        body = {**self.body}
        body['data']['id'] = record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, body, headers=self.headers)
        headers = {**self.headers, 'If-Match': '"12345"'}
        self.app.put_json(record_url, body, headers=headers, status=412)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_patch_sends_update_action(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])

        self.app.patch_json(record_url, self.body, headers=self.headers,
                            status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.UPDATE.value)

    def test_put_sends_update_action_if_record_exists(self):
        body = {**self.body}
        body['data']['id'] = record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, body,
                          headers=self.headers, status=201)

        body['data']['more'] = 'stuff'
        self.app.put_json(record_url, body,
                          headers=self.headers, status=200)

        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.UPDATE.value)

    def test_delete_sends_delete_action(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])

        self.app.delete(record_url, headers=self.headers, status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.DELETE.value)

    def test_collection_delete_sends_delete_action(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)

        self.app.delete(self.collection_url, headers=self.headers, status=200)

        self.assertEqual(len(self.events), 3)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[2].payload['action'],
                         ACTIONS.DELETE.value)

    def test_request_fails_if_notify_fails(self):
        with notif_broken(self.app.app, ResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers, status=500)
        self.assertEqual(len(self.events), 0)

    def test_triggered_on_protected_resource(self):
        app = self.make_app(settings={
            'psilo_write_principals': 'system.Authenticated'
        })
        app.post_json('/psilos', self.body,
                      headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_permissions_are_stripped_from_event_on_protected_resource(self):
        app = self.make_app(settings={
            'psilo_write_principals': 'system.Authenticated'
        })
        resp = app.post_json('/psilos', self.body,
                             headers=self.headers, status=201)
        record = resp.json['data']
        record_url = '/psilos/{}'.format(record['id'])
        app.patch_json(record_url, {"data": {"name": "De barcelona"}},
                       headers=self.headers)
        impacted_records = self.events[-1].impacted_records
        self.assertNotIn('__permissions__', impacted_records[0]['new'])
        self.assertNotIn('__permissions__', impacted_records[0]['old'])


class AfterResourceChangedTest(BaseEventTest, unittest.TestCase):

    subscribed = (AfterResourceChanged,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers)

        self.assertEqual(len(self.events), 0)


class AfterResourceReadTest(BaseEventTest, unittest.TestCase):

    subscribed = (AfterResourceRead,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers)

        self.assertEqual(len(self.events), 0)


class ImpactedRecordsTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged,)

    def test_create_has_new_record_and_no_old_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record = resp.json['data']
        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertNotIn('old', impacted_records[0])
        self.assertEqual(impacted_records[0]['new'], record)

    def test_collection_delete_has_old_and_new_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record1 = resp.json['data']
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record2 = resp.json['data']

        self.app.delete(self.collection_url, headers=self.headers, status=200)

        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 2)
        impacted_records = sorted(impacted_records,
                                  key=lambda x: x['old']['last_modified'])
        self.assertEqual(impacted_records[0]['old'], record1)
        self.assertEqual(impacted_records[1]['old'], record2)
        self.assertEqual(impacted_records[0]['new']['deleted'], True)
        self.assertEqual(impacted_records[1]['new']['deleted'], True)
        self.assertEqual(impacted_records[0]['old']['id'],
                         impacted_records[0]['new']['id'])
        self.assertEqual(impacted_records[1]['old']['id'],
                         impacted_records[1]['new']['id'])
        deleted_ids = {impacted_records[0]['new']['id'],
                       impacted_records[1]['new']['id']}
        self.assertEqual(deleted_ids, {record1['id'], record2['id']})

    def test_update_has_old_and_new_record(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])
        self.app.patch_json(record_url, {'data': {'name': 'en boite'}},
                            headers=self.headers)
        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertEqual(impacted_records[0]['new']['id'], record['id'])
        self.assertEqual(impacted_records[0]['new']['id'],
                         impacted_records[0]['old']['id'])
        self.assertEqual(impacted_records[0]['old']['name'], 'de Paris')
        self.assertEqual(impacted_records[0]['new']['name'], 'en boite')

    def test_delete_has_old_and_new_record_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])
        self.app.delete(record_url, headers=self.headers, status=200)

        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertEqual(impacted_records[0]['old'], record)
        self.assertEqual(impacted_records[0]['new']['id'], record['id'])
        self.assertEqual(impacted_records[0]['new']['deleted'], True)


class BatchEventsTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged, ResourceRead)

    def test_impacted_records_are_merged(self):
        record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        body = {
            "defaults": {
                "method": "PUT",
                "path": record_url
            },
            "requests": [
                {"body": {'data': {'name': 'foo'}}},
                {"body": {'data': {'name': 'bar'}}},
                {"body": {'data': {'name': 'baz'}}},
                {"method": "DELETE"}
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

        create_event = self.events[0]
        self.assertEqual(create_event.payload['action'], 'create')
        self.assertEqual(len(create_event.impacted_records), 1)
        self.assertNotIn('old', create_event.impacted_records[0])
        update_event = self.events[1]
        self.assertEqual(update_event.payload['action'], 'update')
        impacted = update_event.impacted_records
        self.assertEqual(len(impacted), 2)
        self.assertEqual(impacted[0]['old']['name'], 'foo')
        self.assertEqual(impacted[0]['new']['name'], 'bar')
        self.assertEqual(impacted[1]['old']['name'], 'bar')
        self.assertEqual(impacted[1]['new']['name'], 'baz')
        delete_event = self.events[2]
        self.assertEqual(delete_event.payload['action'], 'delete')
        self.assertEqual(len(delete_event.impacted_records), 1)
        self.assertIn('old', delete_event.impacted_records[0])
        self.assertIn('new', delete_event.impacted_records[0])

    def test_one_event_is_sent_per_resource(self):
        body = {
            "defaults": {
                "method": "POST",
                "body": self.body,
            },
            "requests": [
                {"path": '/mushrooms'},
                {"path": '/mushrooms'},
                {"path": '/psilos'},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_parent_id(self):
        # /mushrooms is a UserResource (see testapp.views), which means
        # that parent_id depends on the authenticated user.
        body = {
            "defaults": {
                "path": '/mushrooms',
                "method": "POST",
                "body": self.body
            },
            "requests": [
                {"headers": {"Authorization": "Basic bWF0OjE="}},
                {"headers": {"Authorization": "Basic dG90bzp0dXR1"}},
                {"headers": {"Authorization": "Basic bWF0OjE="}},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        # Two different auth headers, thus two different parent_id:
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_action(self):
        body = {
            "defaults": {
                "path": '/mushrooms',
            },
            "requests": [
                {"method": "POST", "body": self.body},
                {"method": "DELETE"},
                {"method": "GET"},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

    def test_events_are_not_sent_if_subrequest_fails(self):
        patch = mock.patch.object(self.storage,
                                  'delete_all',
                                  side_effect=BackendError('boom'))
        patch.start()
        self.addCleanup(patch.stop)
        request_create = {
            "method": "POST",
            "body": self.body,
        }
        request_delete_all = {
            "method": "DELETE",
            "body": self.body,
        }
        body = {
            "defaults": {
                "path": self.collection_url
            },
            "requests": [request_create, request_delete_all]
        }
        self.app.post_json("/batch", body, headers=self.headers,
                           status=503)
        self.assertEqual(len(self.events), 0)


class CascadingEventsTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged, AfterResourceChanged)

    @classmethod
    def listener(cls, event):
        """Have the first "de Paris" event trigger a follow-up "de New York" event."""
        cls.events.append(event)
        if not isinstance(event, ResourceChanged):
            return
        if len(event.impacted_records) > 1:
            raise ValueError("Too many events {}".format(event.impacted_records))
        # An event without records is impossible.
        assert len(event.impacted_records) == 1
        if event.impacted_records[0]['new']['name'] == 'de Paris':
            new_record = {'name': 'de New York'}
            # Trying to match the Mushroom (i.e. UserResource) parent
            # ID to test event grouping
            parent_id = event.request.prefixed_userid
            notify_resource_event(event.request, parent_id, event.payload['timestamp'], new_record,
                                  ACTIONS.CREATE)

    def test_event_can_trigger_other_event(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        resource_changed_events = [e for e in self.events if isinstance(e, ResourceChanged)]
        self.assertEqual(len(resource_changed_events), 2)
        self.assertEqual(resource_changed_events[0].payload['action'], ACTIONS.CREATE.value)
        self.assertEqual(len(resource_changed_events[0].impacted_records), 1)
        self.assertEqual(resource_changed_events[0].impacted_records[0]['new']['name'],
                         'de Paris')
        self.assertEqual(resource_changed_events[1].payload['action'], ACTIONS.CREATE.value)
        self.assertEqual(len(resource_changed_events[1].impacted_records), 1)
        self.assertEqual(resource_changed_events[1].impacted_records[0]['new']['name'],
                         'de New York')

    def test_cascading_events_are_merged(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        relevant_events = [e for e in self.events if isinstance(e, AfterResourceChanged)]
        self.assertEqual(len(relevant_events), 1)
        merged_event = relevant_events[0]
        self.assertEqual(merged_event.payload['action'], ACTIONS.CREATE.value)
        self.assertEqual(len(merged_event.impacted_records), 2)
        self.assertEqual(merged_event.impacted_records[0]['new']['name'], 'de Paris')
        self.assertEqual(merged_event.impacted_records[1]['new']['name'], 'de New York')


def load_from_config(config, prefix):
    class ClassListener:
        def __call__(self, event):
            pass
    return ClassListener()


@unittest.skipIf(not statsd.statsd_module, "statsd is not installed.")
class StatsDTest(BaseWebTest, unittest.TestCase):

    @classmethod
    def get_app_settings(cls, *args, **kwargs):
        settings = super().get_app_settings(*args, **kwargs)
        if not statsd.statsd_module:
            return settings

        settings['statsd_url'] = 'udp://localhost:8125'
        this_module = 'tests.core.resource.test_events'
        settings['event_listeners'] = 'test'
        settings['event_listeners.test.use'] = this_module
        return settings

    def test_statds_tracks_listeners_execution_duration(self):
        statsd_client = self.app.app.registry.statsd._client
        with mock.patch.object(statsd_client, 'timing') as mocked:
            self.app.post_json(self.collection_url,
                               {"data": {"name": "pouet"}},
                               headers=self.headers)
            timers = set(c[0][0] for c in mocked.call_args_list)
            self.assertIn('listeners.test', timers)



import logging
import re
import functools
from uuid import uuid4

import colander
import venusian

from pyramid import exceptions as pyramid_exceptions
from pyramid.decorator import reify
from pyramid.security import Everyone
from pyramid.httpexceptions import (HTTPNotModified, HTTPPreconditionFailed,
                                    HTTPNotFound, HTTPServiceUnavailable)

from kinto.core import Service
from kinto.core.errors import http_error, raise_invalid, send_alert, ERRORS, request_GET
from kinto.core.events import ACTIONS
from kinto.core.storage import exceptions as storage_exceptions, Filter, Sort, MISSING
from kinto.core.utils import (
    COMPARISON, classname, decode64, encode64, json, find_nested_value,
    dict_subset, recursive_update_dict, apply_json_patch
)

from .model import Model, ShareableModel
from .schema import ResourceSchema, JsonPatchRequestSchema
from .viewset import ViewSet, ShareableViewSet


logger = logging.getLogger(__name__)


def register(depth=1, **kwargs):
    """Ressource class decorator.

    Register the decorated class in the cornice registry.
    Pass all its keyword arguments to the register_resource
    function.
    """
    def wrapped(resource):
        register_resource(resource, depth=depth + 1, **kwargs)
        return resource
    return wrapped


def register_resource(resource_cls, settings=None, viewset=None, depth=1,
                      **kwargs):
    """Register a resource in the cornice registry.

    :param resource_cls:
        The resource class to register.
        It should be a class or have a "name" attribute.

    :param viewset:
        A ViewSet object, which will be used to find out which arguments should
        be appended to the views, and where the views are.

    :param depth:
        A depth offset. It will be used to determine what is the level of depth
        in the call tree. (set to 1 by default.)

    Any additional keyword parameters will be used to override the viewset
    attributes.
    """
    if viewset is None:
        viewset = resource_cls.default_viewset(**kwargs)
    else:
        viewset.update(**kwargs)

    resource_name = viewset.get_name(resource_cls)

    def register_service(endpoint_type, settings):
        """Registers a service in cornice, for the given type.
        """
        path_pattern = getattr(viewset, '{}_path'.format(endpoint_type))
        path_values = {'resource_name': resource_name}
        path = path_pattern.format_map(path_values)

        name = viewset.get_service_name(endpoint_type, resource_cls)

        service = Service(name, path, depth=depth,
                          **viewset.get_service_arguments())

        # Attach viewset and resource to the service for later reference.
        service.viewset = viewset
        service.resource = resource_cls
        service.type = endpoint_type
        # Attach collection and record paths.
        service.collection_path = viewset.collection_path.format_map(path_values)
        service.record_path = (viewset.record_path.format_map(path_values)
                               if viewset.record_path is not None else None)

        methods = getattr(viewset, '{}_methods'.format(endpoint_type))
        for method in methods:
            if not viewset.is_endpoint_enabled(
                    endpoint_type, resource_name, method.lower(), settings):
                continue

            argument_getter = getattr(viewset, '{}_arguments'.format(endpoint_type))
            view_args = argument_getter(resource_cls, method)

            view = viewset.get_view(endpoint_type, method.lower())
            service.add_view(method, view, klass=resource_cls, **view_args)

            # We support JSON-patch on PATCH views. Since the body payload
            # of JSON Patch is not a dict (mapping) but an array, we can't
            # use the same schema as for other PATCH protocols. We add another
            # dedicated view for PATCH, but targetting a different content_type
            # predicate.
            if method.lower() == 'patch':
                view_args['content_type'] = 'application/json-patch+json'
                view_args['schema'] = JsonPatchRequestSchema()
                service.add_view(method, view, klass=resource_cls, **view_args)

        return service

    def callback(context, name, ob):
        # get the callbacks registred by the inner services
        # and call them from here when the @resource classes are being
        # scanned by venusian.
        config = context.config.with_package(info.module)

        # Storage is mandatory for resources.
        if not hasattr(config.registry, 'storage'):
            msg = 'Mandatory storage backend is missing from configuration.'
            raise pyramid_exceptions.ConfigurationError(msg)

        # A service for the list.
        service = register_service('collection', config.registry.settings)
        config.add_cornice_service(service)
        # An optional one for record endpoint.
        if getattr(viewset, 'record_path') is not None:
            service = register_service('record', config.registry.settings)
            config.add_cornice_service(service)

    info = venusian.attach(resource_cls, callback, category='pyramid', depth=depth)
    return callback


class UserResource:
    """Base resource class providing every endpoint.

    Resources inheriting from UserResource are automatically "scoped"
    by user (see get_parent_id()), with the effect that one user
    cannot look at another user's data. This is good for implementing
    sensitive or private information such as accounts.

    However, most resources in Kinto can be shared by different users,
    with different levels of access determined by their
    permissions. Those resources should inherit from
    ShareableResource, below.

    """

    default_viewset = ViewSet
    """Default :class:`kinto.core.resource.viewset.ViewSet` class to use when
    the resource is registered."""

    default_model = Model
    """Default :class:`kinto.core.resource.model.Model` class to use for
    interacting the :mod:`kinto.core.storage` and :mod:`kinto.core.permission`
    backends."""

    schema = ResourceSchema
    """Schema to validate records."""

    def __init__(self, request, context=None):
        self.request = request
        self.context = context
        self.record_id = self.request.matchdict.get('id')
        self.force_patch_update = False

        content_type = str(self.request.headers.get('Content-Type')).lower()
        self._is_json_patch = content_type == 'application/json-patch+json'
        self._is_merge_patch = content_type == 'application/merge-patch+json'

        # Models are isolated by user.
        parent_id = self.get_parent_id(request)

        # Authentication to storage is transmitted as is (cf. cloud_storage).
        auth = request.headers.get('Authorization')

        if not hasattr(self, 'model'):
            self.model = self.default_model(
                storage=request.registry.storage,
                id_generator=self.id_generator,
                collection_id=classname(self),
                parent_id=parent_id,
                auth=auth)

        # Initialize timestamp as soon as possible.
        self.timestamp

    @reify
    def id_generator(self):
        # ID generator by resource name in settings.
        default_id_generator = self.request.registry.id_generators['']
        resource_name = self.request.current_resource_name
        id_generator = self.request.registry.id_generators.get(resource_name,
                                                               default_id_generator)
        return id_generator

    @reify
    def timestamp(self):
        """Return the current collection timestamp.

        :rtype: int
        """
        try:
            return self.model.timestamp()
        except storage_exceptions.BackendError as e:
            is_readonly = self.request.registry.settings['readonly']
            if not is_readonly:
                raise e
            # If the instance is configured to be readonly, and if the
            # collection is empty, the backend will try to bump the timestamp.
            # It fails if the configured db user has not write privileges.
            logger.exception(e)
            error_msg = ('Collection timestamp cannot be written. '
                         'Records endpoint must be hit at least once from a '
                         'writable instance.')
            raise http_error(HTTPServiceUnavailable(),
                             errno=ERRORS.BACKEND,
                             message=error_msg)

    def get_parent_id(self, request):
        """Return the parent_id of the resource with regards to the current
        request.

        :param request:
            The request used to create the resource.

        :rtype: str

        """
        return request.prefixed_userid

    def _get_known_fields(self):
        """Return all the `field` defined in the ressource schema."""
        known_fields = [c.name for c in self.schema().children] + \
                       [self.model.id_field,
                        self.model.modified_field,
                        self.model.deleted_field]
        return known_fields

    def is_known_field(self, field):
        """Return ``True`` if `field` is defined in the resource schema.
        If the resource schema allows unknown fields, this will always return
        ``True``.

        :param str field: Field name
        :rtype: bool

        """
        if self.schema.get_option('preserve_unknown'):
            return True

        known_fields = self._get_known_fields()
        # Test first level only: ``target.data.id`` -> ``target``
        field = field.split('.', 1)[0]
        return field in known_fields

    #
    # End-points
    #

    def collection_get(self):
        """Model ``GET`` endpoint: retrieve multiple records.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and collection not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.
        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters or sorting are invalid.
        """
        self._add_timestamp_header(self.request.response)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified()
        # Collections are considered resources that always exist
        self._raise_412_if_modified(record={})

        headers = self.request.response.headers

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        partial_fields = self._extract_partial_fields()

        filter_fields = [f.field for f in filters]
        include_deleted = self.model.modified_field in filter_fields

        pagination_rules, offset = self._extract_pagination_rules_from_token(
            limit, sorting)

        records, total_records = self.model.get_records(
            filters=filters,
            sorting=sorting,
            limit=limit,
            pagination_rules=pagination_rules,
            include_deleted=include_deleted)

        offset = offset + len(records)
        if limit and len(records) == limit and offset < total_records:
            lastrecord = records[-1]
            next_page = self._next_page_url(sorting, limit, lastrecord, offset)
            headers['Next-Page'] = next_page

        if partial_fields:
            records = [
                dict_subset(record, partial_fields)
                for record in records
            ]

        headers['Total-Records'] = str(total_records)

        return self.postprocess(records)

    def collection_post(self):
        """Model ``POST`` endpoint: create a record.

        If the new record id conflicts against an existing one, the
        posted record is ignored, and the existing record is returned, with
        a ``200`` status.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.process_record`
        """
        new_record = self.request.validated['body'].get('data', {})
        try:
            # Since ``id`` does not belong to schema, it is not in validated
            # data. Must look up in body.
            id_field = self.model.id_field
            new_record[id_field] = _id = self.request.json['data'][id_field]
            self._raise_400_if_invalid_id(_id)
            existing = self._get_record_or_404(_id)
        except (HTTPNotFound, KeyError, ValueError):
            existing = None

        self._raise_412_if_modified(record=existing)

        if existing:
            record = existing
            action = ACTIONS.READ
        else:
            new_record = self.process_record(new_record)
            record = self.model.create_record(new_record)
            self.request.response.status_code = 201
            action = ACTIONS.CREATE

        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(record, action=action)

    def collection_delete(self):
        """Model ``DELETE`` endpoint: delete multiple records.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters are invalid.
        """
        # Collections are considered resources that always exist
        self._raise_412_if_modified(record={})

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        pagination_rules, offset = self._extract_pagination_rules_from_token(limit, sorting)

        records, total_records = self.model.get_records(filters=filters,
                                                        sorting=sorting,
                                                        limit=limit,
                                                        pagination_rules=pagination_rules)
        deleted = self.model.delete_records(filters=filters,
                                            sorting=sorting,
                                            limit=limit,
                                            pagination_rules=pagination_rules)
        if deleted:
            lastrecord = deleted[-1]
            # Get timestamp of the last deleted field
            timestamp = lastrecord[self.model.modified_field]
            self._add_timestamp_header(self.request.response, timestamp=timestamp)

            # Add pagination header
            if limit and len(deleted) == limit and total_records > 1:
                next_page = self._next_page_url(sorting, limit, lastrecord, offset)
                self.request.response.headers['Next-Page'] = next_page
        else:
            self._add_timestamp_header(self.request.response)

        headers = self.request.response.headers
        headers['Total-Records'] = str(total_records)

        action = len(deleted) > 0 and ACTIONS.DELETE or ACTIONS.READ
        return self.postprocess(deleted, action=action, old=records)

    def get(self):
        """Record ``GET`` endpoint: retrieve a record.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and record not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.record_id)
        record = self._get_record_or_404(self.record_id)
        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified(record)
        self._raise_412_if_modified(record)

        partial_fields = self._extract_partial_fields()
        if partial_fields:
            record = dict_subset(record, partial_fields)

        return self.postprocess(record)

    def put(self):
        """Record ``PUT`` endpoint: create or replace the provided record and
        return it.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.

        .. note::

            If ``If-None-Match: *`` request header is provided, the
            ``PUT`` will succeed only if no record exists with this id.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.process_record`.
        """
        self._raise_400_if_invalid_id(self.record_id)
        try:
            existing = self._get_record_or_404(self.record_id)
        except HTTPNotFound:
            existing = None

        self._raise_412_if_modified(record=existing)

        # If `data` is not provided, use existing record (or empty if creation)
        post_record = self.request.validated['body'].get('data', existing) or {}

        record_id = post_record.setdefault(self.model.id_field, self.record_id)
        self._raise_400_if_id_mismatch(record_id, self.record_id)

        new_record = self.process_record(post_record, old=existing)

        if existing:
            record = self.model.update_record(new_record)
        else:
            record = self.model.create_record(new_record)
            self.request.response.status_code = 201

        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        action = existing and ACTIONS.UPDATE or ACTIONS.CREATE
        return self.postprocess(record, action=action, old=existing)

    def patch(self):
        """Record ``PATCH`` endpoint: modify a record and return its
        new version.

        If a request header ``Response-Behavior`` is set to ``light``,
        only the fields whose value was changed are returned.
        If set to ``diff``, only the fields whose value became different than
        the one provided are returned.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.

        .. seealso::
            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.apply_changes` or
            :meth:`kinto.core.resource.UserResource.process_record`.
        """
        self._raise_400_if_invalid_id(self.record_id)
        existing = self._get_record_or_404(self.record_id)
        self._raise_412_if_modified(existing)

        # patch is specified as a list of of operations (RFC 6902)
        if self._is_json_patch:
            requested_changes = self.request.validated['body']
        else:
            # `data` attribute may not be present if only perms are patched.
            body = self.request.validated['body']
            if not body:
                # If no `data` nor `permissions` is provided in patch, reject!
                # XXX: This should happen in schema instead (c.f. ShareableViewSet)
                error_details = {
                    'name': 'data',
                    'description': 'Provide at least one of data or permissions',
                }
                raise_invalid(self.request, **error_details)
            requested_changes = body.get('data', {})

        updated, applied_changes = self.apply_changes(existing,
                                                      requested_changes=requested_changes)

        record_id = updated.setdefault(self.model.id_field,
                                       self.record_id)
        self._raise_400_if_id_mismatch(record_id, self.record_id)

        new_record = self.process_record(updated, old=existing)

        changed_fields = [k for k in applied_changes.keys()
                          if existing.get(k) != new_record.get(k)]

        # Save in storage if necessary.
        if changed_fields or self.force_patch_update:
            new_record = self.model.update_record(new_record)

        else:
            # Behave as if storage would have added `id` and `last_modified`.
            for extra_field in [self.model.modified_field,
                                self.model.id_field]:
                new_record[extra_field] = existing[extra_field]

        # Adjust response according to ``Response-Behavior`` header
        body_behavior = self.request.validated['header'].get('Response-Behavior', 'full')

        if body_behavior.lower() == 'light':
            # Only fields that were changed.
            data = {k: new_record[k] for k in changed_fields}

        elif body_behavior.lower() == 'diff':
            # Only fields that are different from those provided.
            data = {k: new_record[k] for k in changed_fields
                    if applied_changes.get(k) != new_record.get(k)}
        else:
            data = new_record

        timestamp = new_record.get(self.model.modified_field,
                                   existing[self.model.modified_field])
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(data, action=ACTIONS.UPDATE, old=existing)

    def delete(self):
        """Record ``DELETE`` endpoint: delete a record and return it.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.record_id)
        record = self._get_record_or_404(self.record_id)
        self._raise_412_if_modified(record)

        # Retreive the last_modified information from a querystring if present.
        last_modified = self.request.validated['querystring'].get('last_modified')

        # If less or equal than current record. Ignore it.
        if last_modified and last_modified <= record[self.model.modified_field]:
            last_modified = None

        try:
            deleted = self.model.delete_record(record, last_modified=last_modified)
        except storage_exceptions.RecordNotFoundError:
            # Delete might fail if the object was deleted since we
            # fetched it from the storage (ref Kinto/kinto#1407). This
            # is one of a larger class of issues where another request
            # could modify the record between our fetch and our
            # delete, which could e.g. invalidate our precondition
            # checking. Fixing this correctly is a larger
            # problem. However, let's punt on fixing it correctly and
            # just handle this one important case for now (see #1557).
            #
            # Raise a 404 vs. a 409 or 412 because that's what we
            # would have done if the other thread's delete had
            # happened a little earlier. (The client doesn't need to
            # know that we did a bunch of work fetching the existing
            # record for nothing.)
            raise self._404_for_record(self.record_id)

        timestamp = deleted[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(deleted, action=ACTIONS.DELETE, old=record)

    #
    # Data processing
    #

    def process_record(self, new, old=None):
        """Hook for processing records before they reach storage, to introduce
        specific logics on fields for example.

        .. code-block:: python

            def process_record(self, new, old=None):
                new = super().process_record(new, old)
                version = old['version'] if old else 0
                new['version'] = version + 1
                return new

        Or add extra validation based on request:

        .. code-block:: python

            from kinto.core.errors import raise_invalid

            def process_record(self, new, old=None):
                new = super().process_record(new, old)
                if new['browser'] not in request.headers['User-Agent']:
                    raise_invalid(self.request, name='browser', error='Wrong')
                return new

        :param dict new: the validated record to be created or updated.
        :param dict old: the old record to be updated,
            ``None`` for creation endpoints.

        :returns: the processed record.
        :rtype: dict
        """
        modified_field = self.model.modified_field
        new_last_modified = new.get(modified_field)

        # Drop the new last_modified if it is not an integer.
        is_integer = isinstance(new_last_modified, int)
        if not is_integer:
            new.pop(modified_field, None)
            return new

        # Drop the new last_modified if lesser or equal to the old one.
        is_less_or_equal = (old is not None and
                            new_last_modified <= old[modified_field])
        if is_less_or_equal:
            new.pop(modified_field, None)

        return new

    def apply_changes(self, record, requested_changes):
        """Merge `changes` into `record` fields.

        .. note::

            This is used in the context of PATCH only.

        Override this to control field changes at record level, for example:

        .. code-block:: python

            def apply_changes(self, record, requested_changes):
                # Ignore value change if inferior
                if record['position'] > changes.get('position', -1):
                    changes.pop('position', None)
                return super().apply_changes(record, requested_changes)

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if result does not comply with resource schema.

        :returns: the new record with `changes` applied.
        :rtype: tuple
        """
        if self._is_json_patch:
            try:
                applied_changes = apply_json_patch(record, requested_changes)['data']
                updated = {**applied_changes}
            except ValueError as e:
                error_details = {
                    'location': 'body',
                    'description': 'JSON Patch operation failed: {}'.format(e)
                }
                raise_invalid(self.request, **error_details)

        else:
            applied_changes = {**requested_changes}
            updated = {**record}

            # recursive patch and remove field if null attribute is passed (RFC 7396)
            if self._is_merge_patch:
                recursive_update_dict(updated, applied_changes, ignores=(None,))
            else:
                updated.update(**applied_changes)

        for field, value in applied_changes.items():
            has_changed = record.get(field, value) != value
            if self.schema.is_readonly(field) and has_changed:
                error_details = {
                    'name': field,
                    'description': 'Cannot modify {}'.format(field)
                }
                raise_invalid(self.request, **error_details)

        try:
            validated = self.schema().deserialize(updated)
        except colander.Invalid as e:
            # Transform the errors we got from colander into Cornice errors.
            # We could not rely on Service schema because the record should be
            # validated only once the changes are applied
            for field, error in e.asdict().items():  # pragma: no branch
                raise_invalid(self.request, name=field, description=error)

        return validated, applied_changes

    def postprocess(self, result, action=ACTIONS.READ, old=None):
        body = {
            'data': result
        }

        parent_id = self.get_parent_id(self.request)
        self.request.notify_resource_event(parent_id=parent_id,
                                           timestamp=self.timestamp,
                                           data=result,
                                           action=action,
                                           old=old)

        return body

    #
    # Internals
    #

    def _404_for_record(self, record_id):
        details = {
            'id': record_id,
            'resource_name': self.request.current_resource_name
        }
        return http_error(HTTPNotFound(), errno=ERRORS.INVALID_RESOURCE_ID,
                          details=details)

    def _get_record_or_404(self, record_id):
        """Retrieve record from storage and raise ``404 Not found`` if missing.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.
        """
        if self.context and self.context.current_record:
            # Set during authorization. Save a storage hit.
            return self.context.current_record

        try:
            return self.model.get_record(record_id)
        except storage_exceptions.RecordNotFoundError:
            raise self._404_for_record(record_id)

    def _add_timestamp_header(self, response, timestamp=None):
        """Add current timestamp in response headers, when request comes in.

        """
        if timestamp is None:
            timestamp = self.timestamp
        # Pyramid takes care of converting.
        response.last_modified = timestamp / 1000.0
        # Return timestamp as ETag.
        response.headers['ETag'] = '"{}"'.format(timestamp)

    def _add_cache_header(self, response):
        """Add Cache-Control and Expire headers, based a on a setting for the
        current resource.

        Cache headers will be set with anonymous requests only.

        .. note::

            The ``Cache-Control: no-cache`` response header does not prevent
            caching in client. It will indicate the client to revalidate
            the response content on each access. The client will send a
            conditional request to the server and check that a
            ``304 Not modified`` is returned before serving content from cache.
        """
        resource_name = self.context.resource_name if self.context else ''
        setting_key = '{}_cache_expires_seconds'.format(resource_name)
        collection_expires = self.request.registry.settings.get(setting_key)
        is_anonymous = self.request.prefixed_userid is None
        if collection_expires and is_anonymous:
            response.cache_expires(seconds=int(collection_expires))
        else:
            # Since `Expires` response header provides an HTTP data with a
            # resolution in seconds, do not use Pyramid `cache_expires()` in
            # order to omit it.
            response.cache_control.no_cache = True
            response.cache_control.no_store = True

    def _raise_400_if_invalid_id(self, record_id):
        """Raise 400 if specified record id does not match the format excepted
        by storage backends.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        is_string = isinstance(record_id, str)
        if not is_string or not self.model.id_generator.match(record_id):
            error_details = {
                'location': 'path',
                'description': 'Invalid record id'
            }
            raise_invalid(self.request, **error_details)

    def _raise_304_if_not_modified(self, record=None):
        """Raise 304 if current timestamp is inferior to the one specified
        in headers.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified`
        """
        if_none_match = self.request.validated['header'].get('If-None-Match')

        if not if_none_match:
            return

        if if_none_match == '*':
            return

        if record:
            current_timestamp = record[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp == if_none_match:
            response = HTTPNotModified()
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_412_if_modified(self, record=None):
        """Raise 412 if current timestamp is superior to the one
        specified in headers.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed`
        """
        if_match = self.request.validated['header'].get('If-Match')
        if_none_match = self.request.validated['header'].get('If-None-Match')

        # Check if record exists
        record_exists = record is not None

        # If no precondition headers, just ignore
        if not if_match and not if_none_match:
            return

        # If-None-Match: * should always raise if a record exists
        if if_none_match == '*' and record_exists:
            modified_since = -1  # Always raise.

        # If-Match should always raise if a record doesn't exist
        elif if_match and not record_exists:
            modified_since = -1

        # If-Match with ETag value on existing records should compare ETag
        elif if_match and if_match != '*':
            modified_since = if_match

        # If none of the above applies, don't raise
        else:
            return

        if record:
            current_timestamp = record[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp != modified_since:
            error_msg = 'Resource was modified meanwhile'
            details = {'existing': record} if record else {}
            response = http_error(HTTPPreconditionFailed(),
                                  errno=ERRORS.MODIFIED_MEANWHILE,
                                  message=error_msg,
                                  details=details)
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_400_if_id_mismatch(self, new_id, record_id):
        """Raise 400 if the `new_id`, within the request body, does not match
        the `record_id`, obtained from request path.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        if new_id != record_id:
            error_msg = 'Record id does not match existing record'
            error_details = {
                'name': self.model.id_field,
                'description': error_msg
            }
            raise_invalid(self.request, **error_details)

    def _extract_partial_fields(self):
        """Extract the fields to do the projection from QueryString parameters.
        """
        fields = self.request.validated['querystring'].get('_fields')
        if fields:
            root_fields = [f.split('.')[0] for f in fields]
            known_fields = self._get_known_fields()
            invalid_fields = set(root_fields) - set(known_fields)
            preserve_unknown = self.schema.get_option('preserve_unknown')
            if not preserve_unknown and invalid_fields:
                error_msg = 'Fields {} do not exist'.format(','.join(invalid_fields))
                error_details = {
                    'name': 'Invalid _fields parameter',
                    'description': error_msg
                }
                raise_invalid(self.request, **error_details)

            # Since id and last_modified are part of the synchronisation
            # API, force their presence in payloads.
            fields = fields + [self.model.id_field, self.model.modified_field]

        return fields

    def _extract_limit(self):
        """Extract limit value from QueryString parameters."""
        paginate_by = self.request.registry.settings['paginate_by']
        max_fetch_size = self.request.registry.settings['storage_max_fetch_size']
        limit = self.request.validated['querystring'].get('_limit', paginate_by)

        # If limit is higher than paginate_by setting, ignore it.
        if limit and paginate_by:
            limit = min(limit, paginate_by)

        # If limit is higher than what storage can retrieve, ignore it.
        limit = min(limit, max_fetch_size) if limit else max_fetch_size

        return limit

    def _extract_filters(self):
        """Extracts filters from QueryString parameters."""
        queryparams = self.request.validated['querystring']

        filters = []

        for param, value in queryparams.items():
            param = param.strip()

            error_details = {
                'name': param,
                'location': 'querystring',
                'description': 'Invalid value for {}'.format(param)
            }

            # Ignore specific fields
            if param.startswith('_') and param not in ('_since',
                                                       '_to',
                                                       '_before'):
                continue

            # Handle the _since specific filter.
            if param in ('_since', '_to', '_before'):

                if param == '_since':
                    operator = COMPARISON.GT
                else:
                    if param == '_to':
                        message = ('_to is now deprecated, '
                                   'you should use _before instead')
                        url = ('https://kinto.readthedocs.io/en/2.4.0/api/'
                               'resource.html#list-of-available-url-'
                               'parameters')
                        send_alert(self.request, message, url)
                    operator = COMPARISON.LT

                if value == '':
                    raise_invalid(self.request, **error_details)

                filters.append(
                    Filter(self.model.modified_field, value, operator)
                )
                continue

            allKeywords = '|'.join([i.name.lower() for i in COMPARISON])
            m = re.match(r'^('+allKeywords+')_([\w\.]+)$', param)
            if m:
                keyword, field = m.groups()
                operator = getattr(COMPARISON, keyword.upper())
            else:
                operator, field = COMPARISON.EQ, param

            if not self.is_known_field(field):
                error_msg = "Unknown filter field '{}'".format(param)
                error_details['description'] = error_msg
                raise_invalid(self.request, **error_details)

            if operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                all_integers = all([isinstance(v, int)
                                    for v in value])
                all_strings = all([isinstance(v, str)
                                   for v in value])
                has_invalid_value = (
                    (field == self.model.id_field and not all_strings) or
                    (field == self.model.modified_field and not all_integers)
                )
                if has_invalid_value:
                    raise_invalid(self.request, **error_details)

            if field == self.model.modified_field and value == '':
                raise_invalid(self.request, **error_details)

            filters.append(Filter(field, value, operator))

        return filters

    def _extract_sorting(self, limit):
        """Extracts filters from QueryString parameters."""
        specified = self.request.validated['querystring'].get('_sort', [])
        sorting = []
        modified_field_used = self.model.modified_field in specified
        for field in specified:
            field = field.strip()
            m = re.match(r'^([\-+]?)([\w\.]+)$', field)
            if m:
                order, field = m.groups()

                if not self.is_known_field(field):
                    error_details = {
                        'location': 'querystring',
                        'description': "Unknown sort field '{}'".format(field)
                    }
                    raise_invalid(self.request, **error_details)

                direction = -1 if order == '-' else 1
                sorting.append(Sort(field, direction))

        if not modified_field_used:
            # Add a sort by the ``modified_field`` in descending order
            # useful for pagination
            sorting.append(Sort(self.model.modified_field, -1))
        return sorting

    def _build_pagination_rules(self, sorting, last_record, rules=None):
        """Return the list of rules for a given sorting attribute and
        last_record.

        """
        if rules is None:
            rules = []

        rule = []
        next_sorting = sorting[:-1]

        for field, _ in next_sorting:
            rule.append(Filter(field, last_record.get(field, MISSING), COMPARISON.EQ))

        field, direction = sorting[-1]

        if direction == -1:
            rule.append(Filter(field, last_record.get(field, MISSING), COMPARISON.LT))
        else:
            rule.append(Filter(field, last_record.get(field, MISSING), COMPARISON.GT))

        rules.append(rule)

        if len(next_sorting) == 0:
            return rules

        return self._build_pagination_rules(next_sorting, last_record, rules)

    def _extract_pagination_rules_from_token(self, limit, sorting):
        """Get pagination params."""
        token = self.request.validated['querystring'].get('_token', None)
        filters = []
        offset = 0
        if token:
            error_msg = None
            try:
                tokeninfo = json.loads(decode64(token))
                if not isinstance(tokeninfo, dict):
                    raise ValueError()
                last_record = tokeninfo['last_record']
                offset = tokeninfo['offset']
                nonce = tokeninfo['nonce']
            except (ValueError, KeyError, TypeError):
                error_msg = '_token has invalid content'

            # We don't want pagination tokens to be reused several times (#1171).
            # The cache backend is used to keep track of "nonces".
            if self.request.method.lower() == 'delete' and error_msg is None:
                registry = self.request.registry
                deleted = registry.cache.delete(nonce)
                if deleted is None:
                    error_msg = '_token was already used or has expired.'

            if error_msg:
                error_details = {
                    'location': 'querystring',
                    'description': error_msg
                }
                raise_invalid(self.request, **error_details)

            filters = self._build_pagination_rules(sorting, last_record)

        return filters, offset

    def _next_page_url(self, sorting, limit, last_record, offset):
        """Build the Next-Page header from where we stopped."""
        token = self._build_pagination_token(sorting, last_record, offset)

        params = {**request_GET(self.request), '_limit': limit, '_token': token}

        service = self.request.current_service
        next_page_url = self.request.route_url(service.name, _query=params,
                                               **self.request.matchdict)
        return next_page_url

    def _build_pagination_token(self, sorting, last_record, offset):
        """Build a pagination token.

        It is a base64 JSON object with the sorting fields values of
        the last_record.

        """
        nonce = 'pagination-token-{}'.format(uuid4())
        if self.request.method.lower() == 'delete':
            registry = self.request.registry
            validity = registry.settings['pagination_token_validity_seconds']
            registry.cache.set(nonce, '', validity)

        token = {
            'last_record': {},
            'offset': offset,
            'nonce': nonce,
        }

        for field, _ in sorting:
            last_value = find_nested_value(last_record, field, MISSING)
            if last_value is not MISSING:
                token['last_record'][field] = last_value

        return encode64(json.dumps(token))


class ShareableResource(UserResource):
    """Shareable resources allow to set permissions on records, in order to
    share their access or protect their modification.
    """
    default_model = ShareableModel
    default_viewset = ShareableViewSet
    permissions = ('read', 'write')
    """List of allowed permissions names."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # In base resource, PATCH only hit storage if no data has changed.
        # Here, we force update because we add the current principal to
        # the ``write`` ACE.
        self.force_patch_update = True

        # Required by the ShareableModel class.
        self.model.permission = self.request.registry.permission
        if self.request.prefixed_userid is None:
            # The principal of an anonymous is system.Everyone
            self.model.current_principal = Everyone
        else:
            self.model.current_principal = self.request.prefixed_userid
        self.model.prefixed_principals = self.request.prefixed_principals

        if self.context:
            self.model.get_permission_object_id = functools.partial(
                self.context.get_permission_object_id,
                self.request)

    def get_parent_id(self, request):
        """Unlike :class:`kinto.core.resource.UserResource`, records are not
        isolated by user.

        See https://github.com/mozilla-services/cliquet/issues/549

        :returns: A constant empty value.
        """
        return ''

    def _extract_filters(self):
        """Override default filters extraction from QueryString to allow
        partial collection of records.

        XXX: find more elegant approach to add custom filters.
        """
        filters = super()._extract_filters()

        ids = self.context.shared_ids
        if ids is not None:
            filter_by_id = Filter(self.model.id_field, ids, COMPARISON.IN)
            filters.insert(0, filter_by_id)

        return filters

    def _raise_412_if_modified(self, record=None):
        """Do not provide the permissions among the record fields.
        Ref: https://github.com/Kinto/kinto/issues/224
        """
        if record:
            record = {**record}
            record.pop(self.model.permissions_field, None)
        return super()._raise_412_if_modified(record)

    def process_record(self, new, old=None):
        """Read permissions from request body, and in the case of ``PUT`` every
        existing ACE is removed (using empty list).
        """
        new = super().process_record(new, old)

        # patch is specified as a list of of operations (RFC 6902)

        payload = self.request.validated['body']

        if self._is_json_patch:
            permissions = apply_json_patch(old, payload)['permissions']

        elif self._is_merge_patch:
            existing = old or {}
            permissions = existing.get('__permissions__', {})
            recursive_update_dict(permissions,
                                  payload.get('permissions', {}),
                                  ignores=(None,))

        else:
            permissions = {k: v for k, v in payload.get('permissions', {}).items()
                           if v is not None}

        annotated = {**new}

        if permissions:
            is_put = (self.request.method.lower() == 'put')
            if is_put or self._is_merge_patch:
                # Remove every existing ACEs using empty lists.
                for perm in self.permissions:
                    permissions.setdefault(perm, [])
            annotated[self.model.permissions_field] = permissions

        return annotated

    def postprocess(self, result, action=ACTIONS.READ, old=None):
        """Add ``permissions`` attribute in response body.

        In the HTTP API, it was decided that ``permissions`` would reside
        outside the ``data`` attribute.
        """
        body = {}

        if not isinstance(result, list):
            # record endpoint.
            perms = result.pop(self.model.permissions_field, None)
            if perms is not None:
                body['permissions'] = {k: list(p) for k, p in perms.items()}

            if old:
                # Remove permissions from event payload.
                old.pop(self.model.permissions_field, None)

        data = super().postprocess(result, action, old)
        body.update(data)
        return body



import unittest

from pyramid.config import Configurator

from kinto.core.testing import get_user_headers
from kinto.events import ServerFlushed

from ..support import (BaseWebTest,
                       MINIMALIST_BUCKET, MINIMALIST_COLLECTION,
                       MINIMALIST_RECORD)


class FlushViewTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley/records'
    events = []

    def setUp(self):
        super().setUp()

        del self.events[:]

        bucket = {**MINIMALIST_BUCKET}

        self.alice_headers = {**self.headers, **get_user_headers('alice')}

        resp = self.app.get('/', headers=self.alice_headers)
        alice_principal = resp.json['user']['id']
        bucket['permissions'] = {'write': [alice_principal]}

        # Create shared bucket.
        self.app.put_json('/buckets/beers', bucket,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)

        # Records for alice and bob.
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.headers,
                           status=201)
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.alice_headers,
                           status=201)

    def tearDown(self):
        del self.events[:]
        super().tearDown()

    @classmethod
    def make_app(cls, settings=None, config=None):
        settings = cls.get_app_settings(settings)
        config = Configurator(settings=settings)
        config.add_subscriber(cls.listener, ServerFlushed)
        config.commit()
        return super().make_app(settings=settings, config=config)

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault('includes', 'kinto.plugins.flush')
        settings = super().get_app_settings(extras)
        return settings

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

    def test_returns_404_if_not_enabled_in_configuration(self):
        app = self.make_app(settings={'includes': ''})
        app.post('/__flush__', headers=self.headers, status=404)

    def test_removes_every_records_of_everykind(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)

        self.app.post('/__flush__', headers=self.headers, status=202)

        self.app.get(self.collection_url, headers=self.headers, status=403)
        self.app.get(self.collection_url,
                     headers=self.alice_headers,
                     status=403)

    def test_event_triggered_post(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)
        self.app.post('/__flush__', headers=self.headers, status=202)
        self.assertEqual(len(self.events), 1)
        self.assertTrue(isinstance(self.events[0], ServerFlushed))

    def test_flush_returns_json(self):
        response = self.app.post('/__flush__', headers=self.headers, status=202)
        self.assertEquals(response.json, {})

    def test_flush_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('flush_endpoint', capabilities)
        expected = {
            "description": "The __flush__ endpoint can be used to remove "
                           "all data from all backends.",
            "url": "https://kinto.readthedocs.io/en/latest/api/1.x/flush.html"
        }
        self.assertEqual(expected, capabilities['flush_endpoint'])



"""
kinto.core.storage.utils: methods for making it easier to work with kinto storage objects
"""
from kinto.core.storage import Filter
from kinto.core.utils import COMPARISON

BATCH_SIZE = 25


def paginated(storage, *args, sorting, batch_size=BATCH_SIZE, **kwargs):
    """A generator used to access paginated results from storage.get_all.

    :param kwargs: Passed through unchanged to get_all.
    """

    if len(sorting) > 1:
        raise NotImplementedError('FIXME: only supports one-length sorting')  # pragma: nocover
    pagination_direction = COMPARISON.GT if sorting[0].direction > 0 else COMPARISON.LT

    record_pagination = None
    while True:
        (records, _) = storage.get_all(sorting=sorting, limit=batch_size,
                                       pagination_rules=record_pagination,
                                       **kwargs)

        if not records:
            break

        for record in records:
            yield record

        record_pagination = [
            # FIXME: support more than one-length sorting
            [Filter(sorting[0].field, record[sorting[0].field], pagination_direction)]
        ]



import re
import operator
from collections import defaultdict
from collections import abc
import numbers

from kinto.core import utils
from kinto.core.decorators import synchronized
from kinto.core.storage import (
    StorageBase, exceptions,
    DEFAULT_ID_FIELD, DEFAULT_MODIFIED_FIELD, DEFAULT_DELETED_FIELD,
    MISSING)
from kinto.core.utils import (COMPARISON, find_nested_value)

import json
import ujson


def tree():
    return defaultdict(tree)


class MemoryBasedStorage(StorageBase):
    """Abstract storage class, providing basic operations and
    methods for in-memory implementations of sorting and filtering.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    def strip_deleted_record(self, collection_id, parent_id, record,
                             id_field=DEFAULT_ID_FIELD,
                             modified_field=DEFAULT_MODIFIED_FIELD,
                             deleted_field=DEFAULT_DELETED_FIELD):
        """Strip the record of all its fields expect id and timestamp,
        and set the deletion field value (e.g deleted=True)
        """
        deleted = {}
        deleted[id_field] = record[id_field]
        deleted[modified_field] = record[modified_field]
        deleted[deleted_field] = True
        return deleted

    def set_record_timestamp(self, collection_id, parent_id, record,
                             modified_field=DEFAULT_MODIFIED_FIELD,
                             last_modified=None):
        timestamp = self.bump_and_store_timestamp(collection_id, parent_id, record,
                                                  modified_field,
                                                  last_modified=last_modified)
        record[modified_field] = timestamp
        return record

    def extract_record_set(self, records,
                           filters, sorting, id_field, deleted_field,
                           pagination_rules=None, limit=None):
        """Take the list of records and handle filtering, sorting and
        pagination.

        """
        return extract_record_set(records,
                                  filters=filters,
                                  sorting=sorting,
                                  id_field=id_field,
                                  deleted_field=deleted_field,
                                  pagination_rules=pagination_rules,
                                  limit=limit)

    def bump_timestamp(self, collection_timestamp,
                       record, modified_field, last_modified):
        """Timestamp are base on current millisecond.

        .. note ::

            Here it is assumed that if requests from the same user burst in,
            the time will slide into the future. It is not problematic since
            the timestamp notion is opaque, and behaves like a revision number.
        """
        is_specified = (record is not None and
                        modified_field in record or
                        last_modified is not None)
        if is_specified:
            # If there is a timestamp in the new record, try to use it.
            if last_modified is not None:
                current = last_modified
            else:
                current = record[modified_field]

            # If it is equal to current collection timestamp, bump it.
            if current == collection_timestamp:
                collection_timestamp += 1
                current = collection_timestamp
            # If it is superior (future), use it as new collection timestamp.
            elif current > collection_timestamp:
                collection_timestamp = current
            # Else (past), do nothing.

        else:
            # Not specified, use a new one.
            current = utils.msec_time()
            # If two ops in the same msec, bump it.
            if current <= collection_timestamp:
                current = collection_timestamp + 1
            collection_timestamp = current
        return current, collection_timestamp

    def bump_and_store_timestamp(self, collection_id, parent_id, record=None,
                                 modified_field=None, last_modified=None):
        """Use the bump_timestamp to get its next value and store the collection_timestamp.
        """
        raise NotImplementedError


class Storage(MemoryBasedStorage):
    """Storage backend implementation in memory.

    Useful for development or testing purposes, but records are lost after
    each server restart.

    Enable in configuration::

        kinto.storage_backend = kinto.core.storage.memory

    Enable strict json validation before saving (instead of the more lenient
    ujson, see #1238)::

        kinto.storage_strict_json = true
    """
    def __init__(self, *args, readonly=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly = readonly
        self.flush()

    def flush(self, auth=None):
        self._store = tree()
        self._cemetery = tree()
        self._timestamps = defaultdict(dict)

    @synchronized
    def collection_timestamp(self, collection_id, parent_id, auth=None):
        ts = self._timestamps[parent_id].get(collection_id)
        if ts is not None:
            return ts
        if self.readonly:
            error_msg = 'Cannot initialize empty collection timestamp when running in readonly.'
            raise exceptions.BackendError(message=error_msg)
        return self.bump_and_store_timestamp(collection_id, parent_id)

    def bump_and_store_timestamp(self, collection_id, parent_id, record=None,
                                 modified_field=None, last_modified=None):
        """Use the bump_timestamp to get its next value and store the collection_timestamp.
        """
        current_collection_timestamp = self._timestamps[parent_id].get(collection_id, 0)

        current, collection_timestamp = self.bump_timestamp(
            current_collection_timestamp,
            record, modified_field,
            last_modified)
        self._timestamps[parent_id][collection_id] = collection_timestamp

        return current

    @synchronized
    def create(self, collection_id, parent_id, record, id_generator=None,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD, auth=None):
        id_generator = id_generator or self.id_generator
        record = {**record}
        if id_field in record:
            # Raise unicity error if record with same id already exists.
            try:
                existing = self.get(collection_id, parent_id, record[id_field])
                raise exceptions.UnicityError(id_field, existing)
            except exceptions.RecordNotFoundError:
                pass
        else:
            record[id_field] = id_generator()

        self.set_record_timestamp(collection_id, parent_id, record,
                                  modified_field=modified_field)
        _id = record[id_field]
        record = ujson.loads(self.json.dumps(record))
        self._store[parent_id][collection_id][_id] = record
        self._cemetery[parent_id][collection_id].pop(_id, None)
        return record

    @synchronized
    def get(self, collection_id, parent_id, object_id,
            id_field=DEFAULT_ID_FIELD,
            modified_field=DEFAULT_MODIFIED_FIELD,
            auth=None):
        collection = self._store[parent_id][collection_id]
        if object_id not in collection:
            raise exceptions.RecordNotFoundError(object_id)
        return {**collection[object_id]}

    @synchronized
    def update(self, collection_id, parent_id, object_id, record,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD,
               auth=None):
        record = {**record}
        record[id_field] = object_id
        record = ujson.loads(self.json.dumps(record))

        self.set_record_timestamp(collection_id, parent_id, record,
                                  modified_field=modified_field)
        self._store[parent_id][collection_id][object_id] = record
        self._cemetery[parent_id][collection_id].pop(object_id, None)
        return record

    @synchronized
    def delete(self, collection_id, parent_id, object_id,
               id_field=DEFAULT_ID_FIELD, with_deleted=True,
               modified_field=DEFAULT_MODIFIED_FIELD,
               deleted_field=DEFAULT_DELETED_FIELD,
               auth=None, last_modified=None):
        existing = self.get(collection_id, parent_id, object_id)
        # Need to delete the last_modified field of the record.
        del existing[modified_field]

        self.set_record_timestamp(collection_id, parent_id, existing,
                                  modified_field=modified_field,
                                  last_modified=last_modified)
        existing = self.strip_deleted_record(collection_id,
                                             parent_id,
                                             existing)

        # Add to deleted items, remove from store.
        if with_deleted:
            deleted = {**existing}
            self._cemetery[parent_id][collection_id][object_id] = deleted
        self._store[parent_id][collection_id].pop(object_id)
        return existing

    @synchronized
    def purge_deleted(self, collection_id, parent_id, before=None,
                      id_field=DEFAULT_ID_FIELD,
                      modified_field=DEFAULT_MODIFIED_FIELD,
                      auth=None):
        parent_id_match = re.compile(parent_id.replace('*', '.*'))
        by_parent_id = {pid: collections
                        for pid, collections in self._cemetery.items()
                        if parent_id_match.match(pid)}
        num_deleted = 0
        for pid, collections in by_parent_id.items():
            if collection_id is not None:
                collections = {collection_id: collections[collection_id]}
            for collection, colrecords in collections.items():
                if before is None:
                    kept = {}
                else:
                    kept = {key: value for key, value in
                            colrecords.items()
                            if value[modified_field] >= before}
                self._cemetery[pid][collection] = kept
                num_deleted += (len(colrecords) - len(kept))
        return num_deleted

    @synchronized
    def get_all(self, collection_id, parent_id, filters=None, sorting=None,
                pagination_rules=None, limit=None, include_deleted=False,
                id_field=DEFAULT_ID_FIELD,
                modified_field=DEFAULT_MODIFIED_FIELD,
                deleted_field=DEFAULT_DELETED_FIELD,
                auth=None):

        records = _get_objects_by_parent_id(self._store, parent_id, collection_id)

        records, count = self.extract_record_set(records=records,
                                                 filters=filters, sorting=None,
                                                 id_field=id_field, deleted_field=deleted_field)
        deleted = []
        if include_deleted:
            deleted = _get_objects_by_parent_id(self._cemetery, parent_id, collection_id)

        records, count = self.extract_record_set(records=records + deleted,
                                                 filters=filters, sorting=sorting,
                                                 id_field=id_field, deleted_field=deleted_field,
                                                 pagination_rules=pagination_rules, limit=limit)
        return records, count

    @synchronized
    def delete_all(self, collection_id, parent_id, filters=None,
                   sorting=None, pagination_rules=None, limit=None,
                   id_field=DEFAULT_ID_FIELD, with_deleted=True,
                   modified_field=DEFAULT_MODIFIED_FIELD,
                   deleted_field=DEFAULT_DELETED_FIELD,
                   auth=None):
        records = _get_objects_by_parent_id(self._store, parent_id, collection_id, with_meta=True)
        records, count = self.extract_record_set(records=records,
                                                 filters=filters,
                                                 sorting=sorting,
                                                 pagination_rules=pagination_rules, limit=limit,
                                                 id_field=id_field,
                                                 deleted_field=deleted_field)

        deleted = [self.delete(r.pop('__collection_id__'),
                               r.pop('__parent_id__'),
                               r[id_field],
                               id_field=id_field, with_deleted=with_deleted,
                               modified_field=modified_field,
                               deleted_field=deleted_field)
                   for r in records]
        return deleted


def extract_record_set(records, filters, sorting,
                       pagination_rules=None, limit=None,
                       id_field=DEFAULT_ID_FIELD,
                       deleted_field=DEFAULT_DELETED_FIELD):
    """Apply filters, sorting, limit, and pagination rules to the list of
    `records`.

    """
    filtered = list(apply_filters(records, filters or []))
    total_records = len(filtered)

    if pagination_rules:
        paginated = []
        for rule in pagination_rules:
            values = apply_filters(filtered, rule)
            paginated.extend(values)
    else:
        paginated = filtered

    sorted_ = apply_sorting(paginated, sorting or [])

    filtered_deleted = len([r for r in sorted_
                            if r.get(deleted_field) is True])

    if limit:
        sorted_ = list(sorted_)[:limit]

    return sorted_, total_records - filtered_deleted


def canonical_json(record):
    return json.dumps(record, sort_keys=True, separators=(',', ':'))


def apply_filters(records, filters):
    """Filter the specified records, using basic iteration.
    """

    def contains_filtering(record_value, search_term):
        if record_value == MISSING:
            return False
        try:
            search_set = set([canonical_json(v) for v in search_term])
            record_value_set = set([canonical_json(v) for v in record_value])
        except TypeError:
            return False
        return record_value_set.intersection(search_set) == search_set

    def contains_any_filtering(record_value, search_term):
        if record_value == MISSING:
            return False
        try:
            search_set = set([canonical_json(v) for v in search_term])
            record_value_set = set([canonical_json(v) for v in record_value])
        except TypeError:
            return False
        return record_value_set.intersection(search_set)

    operators = {
        COMPARISON.LT: operator.lt,
        COMPARISON.MAX: operator.le,
        COMPARISON.EQ: operator.eq,
        COMPARISON.NOT: operator.ne,
        COMPARISON.MIN: operator.ge,
        COMPARISON.GT: operator.gt,
        COMPARISON.IN: operator.contains,
        COMPARISON.EXCLUDE: lambda x, y: not operator.contains(x, y),
        COMPARISON.LIKE: lambda x, y: re.search(y, x, re.IGNORECASE),
        COMPARISON.CONTAINS: contains_filtering,
        COMPARISON.CONTAINS_ANY: contains_any_filtering,
    }
    for record in records:
        matches = True
        for f in filters:
            right = f.value
            if f.field == DEFAULT_ID_FIELD:
                if isinstance(right, int):
                    right = str(right)

            left = find_nested_value(record, f.field, MISSING)

            if f.operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                right, left = left, right
            elif f.operator == COMPARISON.LIKE:
                # Add implicit start/end wildchars if none is specified.
                if '*' not in right:
                    right = '*{}*'.format(right)
                right = '^{}$'.format(right.replace('*', '.*'))
            elif f.operator in (COMPARISON.LT, COMPARISON.MAX, COMPARISON.EQ,
                                COMPARISON.NOT, COMPARISON.MIN, COMPARISON.GT):
                left = schwartzian_transform(left)
                right = schwartzian_transform(right)

            if f.operator == COMPARISON.HAS:
                matches = left != MISSING if f.value else left == MISSING
            else:
                matches = matches and operators[f.operator](left, right)
        if matches:
            yield record


def schwartzian_transform(value):
    """Decorate a value with a tag that enforces the Postgres sort order.

    The sort order, per https://www.postgresql.org/docs/9.6/static/datatype-json.html, is:

    Object > Array > Boolean > Number > String > Null

    Note that there are more interesting rules for comparing objects
    and arrays but we probably don't need to be that compatible.

    MISSING represents what would be a SQL NULL, which is "bigger"
    than everything else.
    """
    if value is None:
        return (0, value)
    if isinstance(value, str):
        return (1, value)
    if isinstance(value, bool):
        # This has to be before Number, because bools are a subclass
        # of int :(
        return (3, value)
    if isinstance(value, numbers.Number):
        return (2, value)
    if isinstance(value, abc.Sequence):
        return (4, value)
    if isinstance(value, abc.Mapping):
        return (5, value)
    if value is MISSING:
        return (6, value)
    raise ValueError('Unknown value: {}'.format(value))   # pragma: no cover


def apply_sorting(records, sorting):
    """Sort the specified records, using cumulative python sorting.
    """
    result = list(records)

    if not result:
        return result

    def column(record, name):
        return schwartzian_transform(find_nested_value(record, name, default=MISSING))

    for sort in reversed(sorting):
        result = sorted(result,
                        key=lambda r: column(r, sort.field),
                        reverse=(sort.direction < 0))

    return result


def _get_objects_by_parent_id(store, parent_id, collection_id, with_meta=False):
    if parent_id is not None:
        parent_id_match = re.compile('^{}$'.format(parent_id.replace('*', '.*')))
        by_parent_id = {pid: collections
                        for pid, collections in store.items()
                        if parent_id_match.match(pid)}
    else:
        by_parent_id = store[parent_id]

    objects = []
    for pid, collections in by_parent_id.items():
        if collection_id is not None:
            collections = {collection_id: collections[collection_id]}
        for collection, colobjects in collections.items():
            for r in colobjects.values():
                if with_meta:
                    objects.append(dict(__collection_id__=collection,
                                        __parent_id__=pid, **r))
                else:
                    objects.append(r)
    return objects


def load_from_config(config):
    settings = {**config.get_settings()}
    strict = settings.get('storage_strict_json', False)
    return Storage(strict_json=strict)
